"""Tests for coils and coilsets."""

import shutil

import numpy as np
import pytest
import scipy

from desc.backend import jnp
from desc.coils import (
    CoilSet,
    FourierPlanarCoil,
    FourierRZCoil,
    FourierXYZCoil,
    MixedCoilSet,
    SplineXYZCoil,
)
from desc.compute import get_params, get_transforms, rpz2xyz, xyz2rpz, xyz2rpz_vec
from desc.examples import get
from desc.geometry import FourierRZCurve, FourierRZToroidalSurface, FourierXYZCurve
from desc.grid import Grid, LinearGrid
from desc.io import load
from desc.magnetic_fields import SumMagneticField, VerticalMagneticField
from desc.utils import dot


class TestCoil:
    """Tests for singular coil objects."""

    @pytest.mark.unit
    def test_biot_savart_all_coils(self):
        """Test biot-savart implementation against analytic formula."""
        coil_grid = LinearGrid(zeta=100, endpoint=False)

        R = 2
        y = 1
        I = 1e7

        By_true = 1e-7 * 2 * np.pi * R**2 * I / (y**2 + R**2) ** (3 / 2)
        Bz_true = 1e-7 * 2 * np.pi * R**2 * I / (y**2 + R**2) ** (3 / 2)

        B_true_xyz = np.atleast_2d([0, By_true, 0])
        grid_xyz = np.atleast_2d([10, y, 0])
        grid_rpz = xyz2rpz(grid_xyz)
        B_true_rpz_xy = xyz2rpz_vec(B_true_xyz, x=grid_xyz[:, 0], y=grid_xyz[:, 1])
        B_true_rpz_phi = xyz2rpz_vec(B_true_xyz, phi=grid_rpz[:, 1])

        # FourierXYZCoil
        coil = FourierXYZCoil(I)
        transforms = get_transforms(["x", "x_s", "ds"], coil, coil_grid)
        params = get_params(["x", "x_s", "ds"], coil)
        B_xyz = coil.compute_magnetic_field(
            grid_xyz,
            basis="xyz",
            source_grid=coil_grid,
            transforms=transforms,
            params=params,
        )
        B_rpz = coil.compute_magnetic_field(
            grid_rpz, basis="rpz", source_grid=coil_grid
        )
        np.testing.assert_allclose(
            B_true_xyz, B_xyz, rtol=1e-3, atol=1e-10, err_msg="Using FourierXYZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_xy, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using FourierXYZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_phi, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using FourierXYZCoil"
        )

        # SplineXYZCoil
        x = coil.compute("x", grid=coil_grid, basis="xyz")["x"]
        coil = SplineXYZCoil(I, X=x[:, 0], Y=x[:, 1], Z=x[:, 2])
        B_xyz = coil.compute_magnetic_field(
            grid_xyz, basis="xyz", source_grid=coil_grid
        )
        B_rpz = coil.compute_magnetic_field(
            grid_rpz, basis="rpz", source_grid=coil_grid
        )
        np.testing.assert_allclose(
            B_true_xyz, B_xyz, rtol=1e-3, atol=1e-10, err_msg="Using SplineXYZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_xy, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using SplineXYZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_phi, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using SplineXYZCoil"
        )

        # FourierPlanarCoil
        coil = FourierPlanarCoil(I)
        B_xyz = coil.compute_magnetic_field(
            grid_xyz, basis="xyz", source_grid=coil_grid
        )
        B_rpz = coil.compute_magnetic_field(
            grid_rpz, basis="rpz", source_grid=coil_grid
        )
        np.testing.assert_allclose(
            B_true_xyz, B_xyz, rtol=1e-3, atol=1e-10, err_msg="Using FourierPlanarCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_xy,
            B_rpz,
            rtol=1e-3,
            atol=1e-10,
            err_msg="Using FourierPlanarCoil",
        )
        np.testing.assert_allclose(
            B_true_rpz_phi,
            B_rpz,
            rtol=1e-3,
            atol=1e-10,
            err_msg="Using FourierPlanarCoil",
        )

        B_true_xyz = np.atleast_2d([0, 0, Bz_true])
        grid_xyz = np.atleast_2d([0, 0, y])
        grid_rpz = xyz2rpz(grid_xyz)
        B_true_rpz_xy = xyz2rpz_vec(B_true_xyz, x=grid_xyz[:, 0], y=grid_xyz[:, 1])
        B_true_rpz_phi = xyz2rpz_vec(B_true_xyz, phi=grid_rpz[:, 1])

        # FourierRZCoil
        coil = FourierRZCoil(I, R_n=np.array([R]), modes_R=np.array([0]))
        B_xyz = coil.compute_magnetic_field(
            grid_xyz, basis="xyz", source_grid=coil_grid
        )
        B_rpz = coil.compute_magnetic_field(
            grid_rpz, basis="rpz", source_grid=coil_grid
        )
        np.testing.assert_allclose(
            B_true_xyz, B_xyz, rtol=1e-3, atol=1e-10, err_msg="Using FourierRZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_xy, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using FourierRZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_phi, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using FourierRZCoil"
        )

        # FourierRZCoil with NFP>1
        coil = FourierRZCoil(I, R_n=np.array([R]), modes_R=np.array([0]), NFP=2)
        B_xyz = coil.compute_magnetic_field(grid_xyz, basis="xyz", source_grid=None)
        B_rpz = coil.compute_magnetic_field(grid_rpz, basis="rpz", source_grid=None)
        np.testing.assert_allclose(
            B_true_xyz, B_xyz, rtol=1e-3, atol=1e-10, err_msg="Using FourierRZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_xy, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using FourierRZCoil"
        )
        np.testing.assert_allclose(
            B_true_rpz_phi, B_rpz, rtol=1e-3, atol=1e-10, err_msg="Using FourierRZCoil"
        )

    @pytest.mark.unit
    def test_biot_savart_vector_potential_all_coils(self):
        """Test biot-savart vec potential implementation against analytic formula."""
        coil_grid = LinearGrid(zeta=100, endpoint=False)

        R = 2
        y = 1
        I = 1e7

        A_true = np.atleast_2d([0, 0, 0])
        grid_xyz = np.atleast_2d([10, y, 0])
        grid_rpz = xyz2rpz(grid_xyz)

        def test(coil, grid_xyz, grid_rpz):
            A_xyz = coil.compute_magnetic_vector_potential(
                grid_xyz, basis="xyz", source_grid=coil_grid
            )
            A_rpz = coil.compute_magnetic_vector_potential(
                grid_rpz, basis="rpz", source_grid=coil_grid
            )
            np.testing.assert_allclose(
                A_true, A_xyz, rtol=1e-3, atol=1e-10, err_msg=f"Using {coil}"
            )
            np.testing.assert_allclose(
                A_true, A_rpz, rtol=1e-3, atol=1e-10, err_msg=f"Using {coil}"
            )
            np.testing.assert_allclose(
                A_true, A_rpz, rtol=1e-3, atol=1e-10, err_msg=f"Using {coil}"
            )

        # FourierXYZCoil
        coil = FourierXYZCoil(I)
        test(coil, grid_xyz, grid_rpz)

        # SplineXYZCoil
        x = coil.compute("x", grid=coil_grid, basis="xyz")["x"]
        coil = SplineXYZCoil(I, X=x[:, 0], Y=x[:, 1], Z=x[:, 2])
        test(coil, grid_xyz, grid_rpz)

        # FourierPlanarCoil
        coil = FourierPlanarCoil(I)
        test(coil, grid_xyz, grid_rpz)

        grid_xyz = np.atleast_2d([0, 0, y])
        grid_rpz = xyz2rpz(grid_xyz)

        # FourierRZCoil
        coil = FourierRZCoil(I, R_n=np.array([R]), modes_R=np.array([0]))
        test(coil, grid_xyz, grid_rpz)
        # test in a CoilSet
        coil2 = CoilSet(coil)
        test(coil2, grid_xyz, grid_rpz)
        # test in a MixedCoilSet
        coil3 = MixedCoilSet(coil2, coil, check_intersection=False)
        coil3[1].current = 0
        test(coil3, grid_xyz, grid_rpz)

    @pytest.mark.unit
    def test_biot_savart_vector_potential_integral_all_coils(self):
        """Test analytic expression of flux integral for all coils."""
        # taken from analytic benchmark in
        # "A Magnetic Diagnostic Code for 3D Fusion Equilibria", Lazerson 2013
        # find flux for concentric loops of varying radii to a circular coil

        coil_grid = LinearGrid(zeta=1000, endpoint=False)

        R = 1
        I = 1e7

        # analytic eqn for "A_phi" (phi is in dl direction for loop)
        def _A_analytic(r):
            # elliptic integral arguments must be k^2, not k,
            # error in original paper and apparently in Jackson EM book too.
            theta = np.pi / 2
            arg = R**2 + r**2 + 2 * r * R * np.sin(theta)
            term_1_num = 4.0e-7 * I * R
            term_1_den = np.sqrt(arg)
            k_sqd = 4 * r * R * np.sin(theta) / arg
            term_2_num = (2 - k_sqd) * scipy.special.ellipk(
                k_sqd
            ) - 2 * scipy.special.ellipe(k_sqd)
            term_2_den = k_sqd
            return term_1_num * term_2_num / term_1_den / term_2_den

        # we only evaluate it at theta=np.pi/2 (b/c it is in spherical coords)
        rs = np.linspace(0.1, 3, 10, endpoint=True)
        N = 200
        curve_grid = LinearGrid(zeta=N)

        def test(
            coil, grid_xyz, grid_rpz, A_true_rpz, correct_flux, rtol=1e-10, atol=1e-12
        ):
            """Test that we compute the correct flux for the given coil."""
            A_xyz = coil.compute_magnetic_vector_potential(
                grid_xyz, basis="xyz", source_grid=coil_grid
            )
            A_rpz = coil.compute_magnetic_vector_potential(
                grid_rpz, basis="rpz", source_grid=coil_grid
            )
            flux_xyz = jnp.sum(
                dot(A_xyz, curve_data["x_s"], axis=-1) * curve_grid.spacing[:, 2]
            )
            flux_rpz = jnp.sum(
                dot(A_rpz, curve_data_rpz["x_s"], axis=-1) * curve_grid.spacing[:, 2]
            )

            np.testing.assert_allclose(
                correct_flux, flux_xyz, rtol=rtol, err_msg=f"Using {coil}"
            )
            np.testing.assert_allclose(
                correct_flux, flux_rpz, rtol=rtol, err_msg=f"Using {coil}"
            )
            np.testing.assert_allclose(
                A_true_rpz,
                A_rpz,
                rtol=rtol,
                atol=atol,
                err_msg=f"Using {coil}",
            )

        for r in rs:
            # A_phi is constant around the loop (no phi dependence)
            A_true_phi = _A_analytic(r) * np.ones(N)
            A_true_rpz = np.vstack(
                (np.zeros_like(A_true_phi), A_true_phi, np.zeros_like(A_true_phi))
            ).T
            correct_flux = np.sum(r * A_true_phi * 2 * np.pi / N)

            curve = FourierXYZCurve(
                X_n=[-r, 0, 0], Y_n=[0, 0, r], Z_n=[0, 0, 0]
            )  # flux loop to integrate A over

            curve_data = curve.compute(["x", "x_s"], grid=curve_grid, basis="xyz")
            curve_data_rpz = curve.compute(["x", "x_s"], grid=curve_grid, basis="rpz")

            grid_rpz = np.vstack(
                [
                    curve_data_rpz["x"][:, 0],
                    curve_data_rpz["x"][:, 1],
                    curve_data_rpz["x"][:, 2],
                ]
            ).T
            grid_xyz = rpz2xyz(grid_rpz)
            # FourierXYZCoil
            coil = FourierXYZCoil(I, X_n=[-R, 0, 0], Y_n=[0, 0, R], Z_n=[0, 0, 0])
            test(
                coil,
                grid_xyz,
                grid_rpz,
                A_true_rpz,
                correct_flux,
                rtol=1e-8,
                atol=1e-12,
            )

            # SplineXYZCoil
            x = coil.compute("x", grid=coil_grid, basis="xyz")["x"]
            coil = SplineXYZCoil(I, X=x[:, 0], Y=x[:, 1], Z=x[:, 2])
            test(
                coil,
                grid_xyz,
                grid_rpz,
                A_true_rpz,
                correct_flux,
                rtol=1e-4,
                atol=1e-12,
            )

            # FourierPlanarCoil
            coil = FourierPlanarCoil(I, center=[0, 0, 0], normal=[0, 0, -1], r_n=R)
            test(
                coil,
                grid_xyz,
                grid_rpz,
                A_true_rpz,
                correct_flux,
                rtol=1e-8,
                atol=1e-12,
            )

            # FourierRZCoil
            coil = FourierRZCoil(I, R_n=np.array([R]), modes_R=np.array([0]))
            test(
                coil,
                grid_xyz,
                grid_rpz,
                A_true_rpz,
                correct_flux,
                rtol=1e-8,
                atol=1e-12,
            )

    @pytest.mark.unit
    def test_properties(self):
        """Test getting/setting attributes for Coil class."""
        current = 4.34
        coil = FourierPlanarCoil(current)
        assert coil.current == current
        new_current = 3.5
        coil.current = new_current
        assert coil.current == new_current

    @pytest.mark.unit
    def test_SumMagneticField_with_Coil(self):
        """Test SumMagneticField working with Coil and MagneticField objects."""
        R = 2
        y = 1
        I = 1e7
        B_Z = 2  # add constant vertical field of 2T
        By_true = 1e-7 * 2 * np.pi * R**2 * I / (y**2 + R**2) ** (3 / 2)
        B_true = np.array([0, By_true, 2])
        coil = FourierXYZCoil(I)

        field = SumMagneticField(coil, VerticalMagneticField(B_Z))
        B_approx = field.compute_magnetic_field(
            np.array([[10, y, 0], [10, -y, 0]]), basis="xyz", source_grid=100
        )[0]
        np.testing.assert_allclose(B_true, B_approx, rtol=1e-3, atol=1e-10)

    @pytest.mark.unit
    def test_adding_MagneticField_with_Coil_or_CoilSet(self):
        """Test MagneticField plus Coil/CoilSet and vice versa."""
        R = 2
        y = 1
        I = 1e7
        B_Z = 2  # add constant vertical field of 2T
        By_true = 1e-7 * 2 * np.pi * R**2 * I / (y**2 + R**2) ** (3 / 2)
        B_true = np.array([0, By_true, 2])
        coil = FourierXYZCoil(I)
        coilset = CoilSet(coil)
        mixedcoilset = MixedCoilSet(coil)

        field1 = coil + VerticalMagneticField(B_Z)
        field2 = VerticalMagneticField(B_Z) + coil
        # coilset + magnetic field (tests __radd__ of field)
        field3 = coilset + VerticalMagneticField(B_Z)
        field4 = VerticalMagneticField(B_Z) + coilset
        field5 = mixedcoilset + VerticalMagneticField(B_Z)
        field6 = VerticalMagneticField(B_Z) + mixedcoilset

        for i, field in enumerate([field1, field2, field3, field4, field5, field6]):
            B_approx = field.compute_magnetic_field(
                np.array([[10, y, 0], [10, -y, 0]]), basis="xyz", source_grid=100
            )[0]
            np.testing.assert_allclose(
                B_true, B_approx, rtol=1e-3, atol=1e-10, err_msg=f"field {i}"
            )

    @pytest.mark.unit
    def test_convert_type(self):
        """Test conversions between coil representations."""
        s = np.linspace(0, 2 * np.pi, 100, endpoint=False)
        coil1 = FourierRZCoil(1e6, [0, 10, 1], [0, 0, 0])
        coil2 = coil1.to_FourierXYZ(s=s)
        coil3 = coil1.to_SplineXYZ(knots=s)
        coil4 = coil1.to_FourierRZ(N=coil1.N)
        coil5 = coil1.to_FourierPlanar(N=10, basis="rpz")

        grid = LinearGrid(zeta=s)
        x1 = coil1.compute("x", grid=grid, basis="xyz")["x"]
        x2 = coil2.compute("x", grid=grid, basis="xyz")["x"]
        x3 = coil3.compute("x", grid=grid, basis="xyz")["x"]
        x4 = coil4.compute("x", grid=grid, basis="xyz")["x"]
        zeta = np.arctan2(  # zeta = polar angle for planar coil for same points
            x1[:, 1] - coil5.center[1],
            x1[:, 0] - coil5.center[0],
        )  # use Grid instead of LinearGrid to prevent node sorting
        grid_planar = Grid(np.array([np.zeros_like(zeta), np.zeros_like(zeta), zeta]).T)
        x5 = coil5.compute("x", grid=grid_planar, basis="xyz")["x"]

        B1 = coil1.compute_magnetic_field(
            np.zeros((1, 3)), source_grid=grid, basis="xyz"
        )
        B2 = coil2.compute_magnetic_field(
            np.zeros((1, 3)), source_grid=grid, basis="xyz"
        )
        B3 = coil3.compute_magnetic_field(
            np.zeros((1, 3)), source_grid=grid, basis="xyz"
        )
        B4 = coil4.compute_magnetic_field(
            np.zeros((1, 3)), source_grid=grid, basis="xyz"
        )
        B5 = coil5.compute_magnetic_field(
            np.zeros((1, 3)), source_grid=grid, basis="xyz"
        )

        np.testing.assert_allclose(x1, x2, atol=1e-12)
        np.testing.assert_allclose(x1, x3, atol=1e-12)
        np.testing.assert_allclose(x1, x4, atol=1e-12)
        np.testing.assert_allclose(x1, x5, atol=1e-12)
        np.testing.assert_allclose(B1, B2, rtol=1e-8, atol=1e-8)
        np.testing.assert_allclose(B1, B3, rtol=1e-3, atol=1e-8)
        np.testing.assert_allclose(B1, B4, rtol=1e-8, atol=1e-8)
        np.testing.assert_allclose(B1, B5, rtol=1e-6, atol=1e-7)


class TestCoilSet:
    """Tests for sets of multiple coils."""

    @pytest.mark.unit
    def test_linspaced_linear(self):
        """Field from straight solenoid."""
        R = 10
        z = np.linspace(0, 10, 10)
        I = 1e7
        n = 10
        Bz_true = np.sum(1e-7 * 2 * np.pi * R**2 * I / (z**2 + R**2) ** (3 / 2))
        B_true = np.array([0, 0, Bz_true])
        coil = FourierRZCoil(0.1)
        coils = CoilSet.linspaced_linear(
            coil, displacement=[0, 0, 10], n=n, endpoint=True
        )
        coils.current = I
        np.testing.assert_allclose(coils.current, I)
        B_approx = coils.compute_magnetic_field(
            [0, 0, z[-1]], basis="xyz", source_grid=32
        )[0]
        np.testing.assert_allclose(B_true, B_approx, rtol=1e-3, atol=1e-10)

    @pytest.mark.unit
    def test_linspaced_angular(self):
        """Field from uniform toroidal solenoid."""
        R = 10
        N = 50
        I = 1e7
        Bp_true = np.sum(1e-7 * 4 * np.pi * N * I / 2 / np.pi / R)
        B_true = np.array([0, Bp_true, 0])
        coil = FourierPlanarCoil()
        coil.current = I
        coils = CoilSet.linspaced_angular(coil, n=N)
        grid = LinearGrid(N=32, endpoint=False)
        transforms = get_transforms(["x", "x_s", "ds"], coil, grid=grid)
        B_approx = coils.compute_magnetic_field(
            [10, 0, 0], basis="rpz", source_grid=grid, transforms=transforms
        )[0]
        np.testing.assert_allclose(B_true, B_approx, rtol=1e-3, atol=1e-10)

        surf = FourierRZToroidalSurface(
            R_lmn=np.array([10, 0.1]),
            Z_lmn=np.array([-0.1]),
            modes_R=np.array([[0, 0], [1, 0]]),
            modes_Z=np.array([[-1, 0]]),
        )

        B_normal, _ = coils.compute_Bnormal(surf)
        np.testing.assert_allclose(B_normal, 0, atol=1e-9)

    @pytest.mark.unit
    def test_from_symmetry(self):
        """Same toroidal solenoid field, but different construction."""
        R = 10
        N = 48
        I = 1e7
        Bp_true = np.sum(1e-7 * 4 * np.pi * N * I / 2 / np.pi / R)
        B_true = np.array([0, Bp_true, 0])
        coil = FourierPlanarCoil(I)
        coils = CoilSet.linspaced_angular(coil, angle=np.pi / 2, n=N // 4)
        coils = MixedCoilSet.from_symmetry(coils, NFP=4)
        grid = LinearGrid(N=32, endpoint=False)
        transforms = get_transforms(["x", "x_s", "ds"], coil, grid=grid)
        B_approx = coils.compute_magnetic_field(
            [10, 0, 0], basis="rpz", source_grid=grid, transforms=transforms
        )[0]
        np.testing.assert_allclose(B_true, B_approx, rtol=1e-3, atol=1e-10)

        # with stellarator symmetry
        NFP = 4
        coil = FourierXYZCoil()
        coil.rotate(angle=np.pi / N)
        coils = CoilSet.linspaced_angular(
            coil, I, [0, 0, 1], np.pi / NFP, N // NFP // 2
        )
        coils2 = MixedCoilSet.from_symmetry(coils, NFP, True)
        B_approx = coils2.compute_magnetic_field(
            [10, 0, 0], basis="rpz", source_grid=32
        )[0]
        np.testing.assert_allclose(B_true, B_approx, rtol=1e-3, atol=1e-10)

        # With a MixedCoilSet as the base coils and only rotation
        coil = FourierPlanarCoil(I)
        coils = [coil] + [FourierXYZCoil(I) for i in range(N // 4 - 1)]
        for i, c in enumerate(coils[1:]):
            c.rotate(angle=2 * np.pi / N * (i + 1))
        coils = MixedCoilSet.from_symmetry(coils, NFP=4)
        grid = LinearGrid(N=32, endpoint=False)
        transforms = get_transforms(["x", "x_s", "ds"], coil, grid=grid)
        B_approx = coils.compute_magnetic_field(
            [10, 0, 0], basis="rpz", source_grid=grid
        )[0]
        np.testing.assert_allclose(B_true, B_approx, rtol=1e-3, atol=1e-10)

    @pytest.mark.unit
    def test_is_self_intersecting_warnings(self):
        """Test warning in from_symmetry for self-intersection."""
        N = 40
        # tilt coils so they cross the symmetry plane
        # and the resulting coils are self-intersecting
        coil = FourierPlanarCoil(normal=[1e-4, 1, 3])
        coils_list_sym = [coil.copy()] + [coil.copy() for i in range(N // 8 - 1)]

        for i, c in enumerate(coils_list_sym[1:]):
            c.rotate(angle=2 * np.pi / N * (i + 1))

        # test the warning for self-intersecting coils, as two
        #  of the coils in each field period lie nearly
        # in the same physical space (intersecting at 2 points) after reflection
        with pytest.warns(UserWarning) as warninfo:
            _ = CoilSet.from_symmetry(coils_list_sym, NFP=4, sym=True)
        assert "nearly intersecting" in str(warninfo[0].message)

    @pytest.mark.unit
    def test_properties(self):
        """Test getting/setting of CoilSet attributes."""
        coil = FourierPlanarCoil()
        coils = CoilSet.linspaced_linear(coil, n=4, displacement=[0, 2, 0])
        data = coils.compute(
            [
                "x",
                "curvature",
                "torsion",
                "frenet_tangent",
                "frenet_normal",
                "frenet_binormal",
            ],
            grid=0,
            basis="xyz",
        )
        np.testing.assert_allclose(
            [dat["x"] for dat in data],
            np.array(
                [
                    [12, 0, 0],
                    [12, 0.5, 0],
                    [12, 1, 0],
                    [12, 1.5, 0],
                ]
            ).reshape((4, 1, 3)),
        )
        np.testing.assert_allclose([dat["curvature"] for dat in data], 1 / 2)
        np.testing.assert_allclose([dat["torsion"] for dat in data], 0)
        T = [dat["frenet_tangent"] for dat in data]
        N = [dat["frenet_normal"] for dat in data]
        B = [dat["frenet_binormal"] for dat in data]
        np.testing.assert_allclose(
            T,
            np.array(
                [
                    [0, 0, -1],
                    [0, 0, -1],
                    [0, 0, -1],
                    [0, 0, -1],
                ]
            ).reshape((4, 1, 3)),
            atol=1e-12,
        )
        np.testing.assert_allclose(
            N,
            np.array(
                [
                    [-1, 0, 0],
                    [-1, 0, 0],
                    [-1, 0, 0],
                    [-1, 0, 0],
                ]
            ).reshape((4, 1, 3)),
            atol=1e-12,
        )
        np.testing.assert_allclose(
            B,
            np.array(
                [
                    [0, 1, 0],
                    [0, 1, 0],
                    [0, 1, 0],
                    [0, 1, 0],
                ]
            ).reshape((4, 1, 3)),
            atol=1e-12,
        )
        data = coils.compute("length", grid=32)
        np.testing.assert_allclose([dat["length"] for dat in data], 2 * 2 * np.pi)
        coils.translate([1, 1, 1])
        data = coils.compute("length", grid=32)
        np.testing.assert_allclose([dat["length"] for dat in data], 2 * 2 * np.pi)
        coils.flip([1, 0, 0])
        data = coils.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"],
            grid=0,
            basis="xyz",
        )
        T = [dat["frenet_tangent"] for dat in data]
        N = [dat["frenet_normal"] for dat in data]
        B = [dat["frenet_binormal"] for dat in data]
        np.testing.assert_allclose(
            T,
            np.array(
                [
                    [0, 0, -1],
                    [0, 0, -1],
                    [0, 0, -1],
                    [0, 0, -1],
                ]
            ).reshape((4, 1, 3)),
            atol=1e-12,
        )
        np.testing.assert_allclose(
            N,
            np.array(
                [
                    [1, 0, 0],
                    [1, 0, 0],
                    [1, 0, 0],
                    [1, 0, 0],
                ]
            ).reshape((4, 1, 3)),
            atol=1e-12,
        )
        np.testing.assert_allclose(
            B,
            np.array(
                [
                    [0, 1, 0],
                    [0, 1, 0],
                    [0, 1, 0],
                    [0, 1, 0],
                ]
            ).reshape((4, 1, 3)),
            atol=1e-12,
        )

    @pytest.mark.unit
    def test_dunder_methods(self):
        """Test methods for combining and calling CoilSet objects."""
        coil1 = FourierXYZCoil()
        coils1 = MixedCoilSet.from_symmetry(coil1, NFP=4)
        coil2 = FourierPlanarCoil(center=[100, 0, 0])
        coils2 = coils1 + [coil2]
        assert coils2[-1] is coil2
        with pytest.warns(UserWarning, match="nearly intersecting"):
            coils2 = coils1 + MixedCoilSet([coil2, coil2], check_intersection=False)
        assert coils2[-1] is coil2

        with pytest.raises(TypeError):
            _ = coils1 + FourierRZCurve()

        with pytest.raises(TypeError):
            coils1[-1] = FourierRZCurve()

        coils1[-1] = coil2
        assert coils1[-1] is coil2

        coils1.insert(-1, coil2)
        with pytest.raises(TypeError):
            coils1.insert(-1, FourierRZCurve())

        assert len(coils1) == 5

        assert coils1[-1] is coil2
        assert coils1[-2] is coil2

        s = coils1[-2:]
        assert s[-1] is coil2

        del coils1[-2]
        assert len(coils1) == 4
        assert coils1[-1] is coil2
        assert coils1[-2].__class__ is coil1.__class__

        coils2 = CoilSet.linspaced_angular(coil1)
        assert coils2[0].equiv(coil1) and not (coils2[0] is coil1)
        coils2[0] = coil1
        assert coils2[0] is coil1
        with pytest.raises(TypeError):
            coils2[1] = coil2
        with pytest.raises(TypeError):
            coils2.insert(4, coil2)

    @pytest.mark.unit
    def test_convert_type(self):
        """Test converting coilsets between different representations."""
        grid = LinearGrid(N=20)
        coil = FourierRZCoil(1e6, [0, 10, 1], [0, 0, 0])

        # MixedCoilSet
        coils1 = MixedCoilSet.linspaced_linear(coil, displacement=[0, 0, 2.5], n=4)
        coils2 = coils1.to_SplineXYZ(grid=grid, check_intersection=False)
        coils3 = coils1.to_FourierXYZ(grid=grid, check_intersection=False)
        coils4 = coils1.to_FourierPlanar(grid=grid, check_intersection=False)
        assert isinstance(coils2, MixedCoilSet)
        assert isinstance(coils3, MixedCoilSet)
        assert isinstance(coils4, MixedCoilSet)
        assert all(isinstance(coil, SplineXYZCoil) for coil in coils2)
        assert all(isinstance(coil, FourierXYZCoil) for coil in coils3)
        assert all(isinstance(coil, FourierPlanarCoil) for coil in coils4)
        x1 = coils1.compute("x", grid=grid, basis="xyz")
        x2 = coils2.compute("x", grid=grid, basis="xyz")
        x3 = coils3.compute("x", grid=grid, basis="xyz")
        zeta = np.arctan2(  # zeta = polar angle for planar coil for same points
            x1[0]["x"][:, 1] - coils4[0].center[1],
            x1[0]["x"][:, 0] - coils4[0].center[0],
        )  # use Grid instead of LinearGrid to prevent node sorting
        grid_planar = Grid(np.array([np.zeros_like(zeta), np.zeros_like(zeta), zeta]).T)
        x4 = coils4.compute("x", grid=grid_planar, basis="xyz")
        np.testing.assert_allclose(
            [xi["x"] for xi in x1], [xi["x"] for xi in x2], atol=1e-12
        )
        np.testing.assert_allclose(
            [xi["x"] for xi in x1], [xi["x"] for xi in x3], atol=1e-12
        )
        np.testing.assert_allclose(
            [xi["x"] for xi in x1], [xi["x"] for xi in x4], atol=1e-12
        )
        B1 = coils1.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        B2 = coils2.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        B3 = coils3.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        B4 = coils4.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        np.testing.assert_allclose(B1, B2, rtol=1e-2)
        np.testing.assert_allclose(B1, B3, rtol=1e-2)
        np.testing.assert_allclose(B1, B4, rtol=1e-2)

        # CoilSet
        coil = coils3[0]  # FourierXYZCoil
        coils5 = CoilSet.linspaced_linear(coil, displacement=[0, 0, 3.5], n=6)
        coils6 = coils5.to_SplineXYZ(grid=grid, check_intersection=False)
        coils7 = coils5.to_FourierRZ(grid=grid, check_intersection=False)
        coils8 = coils5.to_FourierPlanar(grid=grid, check_intersection=False)
        assert isinstance(coils6, CoilSet)
        assert isinstance(coils7, CoilSet)
        assert isinstance(coils8, CoilSet)
        assert all(isinstance(coil, SplineXYZCoil) for coil in coils6)
        assert all(isinstance(coil, FourierRZCoil) for coil in coils7)
        assert all(isinstance(coil, FourierPlanarCoil) for coil in coils8)
        x5 = coils5.compute("x", grid=grid, basis="xyz")
        x6 = coils6.compute("x", grid=grid, basis="xyz")
        x7 = coils7.compute("x", grid=grid, basis="xyz")
        zeta = np.arctan2(  # zeta = polar angle for planar coil for same points
            x5[0]["x"][:, 1] - coils8[0].center[1],
            x5[0]["x"][:, 0] - coils8[0].center[0],
        )  # use Grid instead of LinearGrid to prevent node sorting
        grid_planar = Grid(np.array([np.zeros_like(zeta), np.zeros_like(zeta), zeta]).T)
        x8 = coils8.compute("x", grid=grid_planar, basis="xyz")
        np.testing.assert_allclose(
            [xi["x"] for xi in x5], [xi["x"] for xi in x6], atol=1e-12
        )
        np.testing.assert_allclose(
            [xi["x"] for xi in x5], [xi["x"] for xi in x7], atol=1e-12
        )
        np.testing.assert_allclose(
            [xi["x"] for xi in x5], [xi["x"] for xi in x8], atol=1e-12
        )
        B5 = coils5.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        B6 = coils6.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        B7 = coils6.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        B8 = coils6.compute_magnetic_field(np.array([[5, 2, 1]]), source_grid=grid)
        np.testing.assert_allclose(B5, B6, rtol=1e-2)
        np.testing.assert_allclose(B5, B7, rtol=1e-2)
        np.testing.assert_allclose(B5, B8, rtol=1e-2)


@pytest.mark.unit
def test_symmetry_position(DummyCoilSet):
    """Tests that compute position is correct from symmetry."""
    # same coil sets with vs without symmetry
    coilset_sym = load(
        load_from=str(DummyCoilSet["output_path_sym"]), file_format="hdf5"
    )
    coilset_asym = load(
        load_from=str(DummyCoilSet["output_path_asym"]), file_format="hdf5"
    )
    coilset_mixed = MixedCoilSet(*coilset_asym)
    grid = LinearGrid(N=30)

    # check that positions of CoilSets are the same with xyz basis
    x_sym = coilset_sym._compute_position(basis="xyz", grid=grid)
    x_asym = coilset_asym._compute_position(basis="xyz", grid=grid)
    x_mixed = coilset_mixed._compute_position(basis="xyz", grid=grid)

    np.testing.assert_allclose(x_sym, x_asym)
    np.testing.assert_allclose(x_sym, x_mixed)

    # check that positions of CoilSets are the same with rpz basis
    x_sym = coilset_sym._compute_position(basis="rpz", grid=grid)
    x_asym = coilset_asym._compute_position(basis="rpz", grid=grid)
    x_mixed = coilset_mixed._compute_position(basis="rpz", grid=grid)

    np.testing.assert_allclose(x_sym, x_asym)
    np.testing.assert_allclose(x_sym, x_mixed)


@pytest.mark.unit
def test_symmetry_magnetic_field(DummyCoilSet):
    """Tests that compute magnetic field is correct from symmetry."""
    # same coil sets with vs without symmetry
    eq = get("precise_QH")
    coilset_sym = load(
        load_from=str(DummyCoilSet["output_path_sym"]), file_format="hdf5"
    )
    coilset_asym = load(
        load_from=str(DummyCoilSet["output_path_asym"]), file_format="hdf5"
    )

    # test that both coil sets compute the same field on the plasma surface
    grid = LinearGrid(rho=[1.0], M=eq.M_grid, N=eq.N_grid, NFP=1, sym=False)
    with pytest.warns(UserWarning):  # because eq.NFP != grid.NFP
        data = eq.compute(["phi", "R", "X", "Y", "Z"], grid)

    # test in (R, phi, Z) coordinates
    nodes_rpz = np.array([data["R"], data["phi"], data["Z"]]).T
    B_sym_rpz = coilset_sym.compute_magnetic_field(nodes_rpz, basis="rpz")
    B_asym_rpz = coilset_asym.compute_magnetic_field(nodes_rpz, basis="rpz")
    np.testing.assert_allclose(B_sym_rpz, B_asym_rpz, atol=1e-14)

    # test in (X, Y, Z) coordinates
    nodes_xyz = np.array([data["X"], data["Y"], data["Z"]]).T
    B_sym_xyz = coilset_sym.compute_magnetic_field(nodes_xyz, basis="xyz")
    B_asym_xyz = coilset_asym.compute_magnetic_field(nodes_xyz, basis="xyz")
    np.testing.assert_allclose(B_sym_xyz, B_asym_xyz, atol=1e-14)


@pytest.mark.unit
def test_load_and_save_makegrid_coils(tmpdir_factory):
    """Test loading in and saving CoilSets from MAKEGRID format files."""
    Ncoils = 22
    input_path = f"./tests/inputs/coils.MAKEGRID_format_{Ncoils}_coils"
    tmpdir = tmpdir_factory.mktemp("coil_files")
    tmp_path = tmpdir.join(f"coils.MAKEGRID_format_{Ncoils}_coils")
    shutil.copyfile(input_path, tmp_path)

    coilset = CoilSet.from_makegrid_coilfile(str(tmp_path))
    assert len(coilset) == Ncoils  # correct number of coils

    path = tmpdir.join("coils.MAKEGRID_format_desc")
    coilset.save_in_makegrid_format(
        str(path), grid=LinearGrid(zeta=coilset[0].knots, theta=0, endpoint=True)
    )

    coilset2 = CoilSet.from_makegrid_coilfile(str(path))

    grid = LinearGrid(N=200, endpoint=False)

    # check values at saved points, ensure they match
    for i, (c1, c2) in enumerate(zip(coilset, coilset2)):
        # make sure knots are exactly the same
        np.testing.assert_allclose(c1.knots, c2.knots, err_msg=f"Coil {i}")

        grid = LinearGrid(zeta=coilset2[0].knots, endpoint=False)
        coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
        X1 = coords1[:, 0]
        Y1 = coords1[:, 1]
        Z1 = coords1[:, 2]

        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
        X2 = coords2[:, 0]
        Y2 = coords2[:, 1]
        Z2 = coords2[:, 2]

        np.testing.assert_allclose(c1.current, c2.current, err_msg=f"Coil {i}")
        np.testing.assert_allclose(X1, X2, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Y1, Y2, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Z1, Z2, atol=2e-7, err_msg=f"Coil {i}")

    # check magnetic field from both, check that matches
    grid = LinearGrid(N=200, endpoint=False)
    B1 = coilset.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )
    B2 = coilset2.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )

    np.testing.assert_allclose(B1, B2, atol=1e-7)


@pytest.mark.unit
def test_load_and_save_makegrid_coils_diff_length_of_knots(tmpdir_factory):
    """Test loading and saving coils from MAKEGRID file that are not uniform length."""
    Ncoils = 2
    input_path = f"./tests/inputs/coils.MAKEGRID_format_{Ncoils}_coils_diff_lengths"
    tmpdir = tmpdir_factory.mktemp("coil_files")
    tmp_path = tmpdir.join(f"coils.MAKEGRID_format_{Ncoils}_coils")
    shutil.copyfile(input_path, tmp_path)

    with pytest.raises(ValueError, match="CoilSet"):
        coilset = CoilSet.from_makegrid_coilfile(str(tmp_path))
    coilset = MixedCoilSet.from_makegrid_coilfile(str(tmp_path), ignore_groups=True)
    assert len(coilset) == Ncoils  # correct number of coils
    # if the coils are not all the same number of knots, then making a CoilSet
    # will fail (as each underyling coil must have same length knots),
    # instead the function should make a MixedCoilSet
    assert isinstance(coilset, MixedCoilSet)

    path = tmpdir.join("coils.MAKEGRID_format_desc")
    # save using the default grids
    coilset.save_in_makegrid_format(str(path))

    coilset2 = MixedCoilSet.from_makegrid_coilfile(str(path))
    assert isinstance(coilset2, MixedCoilSet)

    grid = LinearGrid(N=50, endpoint=False)

    # check values, ensure they are close
    for i, (c1, c2) in enumerate(zip(coilset, coilset2)):

        coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
        X1 = coords1[:, 0]
        Y1 = coords1[:, 1]
        Z1 = coords1[:, 2]

        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
        X2 = coords2[:, 0]
        Y2 = coords2[:, 1]
        Z2 = coords2[:, 2]
        # knots are not the exact same, so these points will be close but not the
        # same.
        np.testing.assert_allclose(c1.current, c2.current, err_msg=f"Coil {i}")
        np.testing.assert_allclose(X1, X2, rtol=2e-5, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Y1, Y2, rtol=2e-5, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Z1, Z2, atol=2e-7, rtol=1e-3, err_msg=f"Coil {i}")

    # check magnetic field from both, check that matches
    grid = LinearGrid(N=200, endpoint=False)
    B1 = coilset.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )
    B2 = coilset2.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )

    np.testing.assert_allclose(B1, B2, atol=1e-7)


@pytest.mark.unit
def test_load_and_save_makegrid_coils_groups(tmpdir_factory):
    """Test loading and saving CoilSets from MAKEGRID format files with coilgroups."""
    Ncoils_per_group = 2
    coilgroups = ["groupone", "grouptwo"]
    input_path = "./tests/inputs/coils.MAKEGRID_format_two_groups"
    tmpdir = tmpdir_factory.mktemp("coil_files")
    tmp_path = tmpdir.join("coils.MAKEGRID_format_two_groups")
    shutil.copyfile(input_path, tmp_path)

    coilset = MixedCoilSet.from_makegrid_coilfile(str(tmp_path))
    assert len(coilset) == len(coilgroups)  # correct number of coils
    for i, (coils, groupname) in enumerate(zip(coilset, coilgroups)):
        assert len(coils) == Ncoils_per_group
        assert groupname in coils.name
        assert str(i + 1) in coils.name  # make sure the correct number is in the name
    path = tmpdir.join("coils.MAKEGRID_format_groups_desc")

    coilset.save_in_makegrid_format(
        str(path), grid=LinearGrid(zeta=coilset[0][0].knots, theta=0, endpoint=True)
    )
    coilset2 = MixedCoilSet.from_makegrid_coilfile(str(path), ignore_groups=False)
    # also compare to flattened
    coilset_flat = MixedCoilSet.from_makegrid_coilfile(str(path), ignore_groups=True)
    assert len(coilset_flat) == Ncoils_per_group * len(coilgroups)

    assert len(coilset2) == len(coilgroups)  # correct number of coils groups
    for i, (coils, groupname) in enumerate(zip(coilset2, coilgroups)):
        assert len(coils) == Ncoils_per_group
        assert groupname in coils.name
        assert str(i + 1) in coils.name  # make sure the correct number is in the name

    grid = LinearGrid(zeta=coilset[0][0].knots, endpoint=False)

    # check values at saved points, ensure they match
    for i, (cs1, cs2) in enumerate(zip(coilset, coilset2)):
        for j, (c1, c2) in enumerate(zip(cs1, cs2)):
            c3 = coilset_flat[2 * i + j]
            print(c1)
            print(c2)
            print(c3)
            # make sure knots are exactly the same
            np.testing.assert_allclose(
                c1.knots, c2.knots, err_msg=f"CoilSet {i} Coil {j}"
            )
            np.testing.assert_allclose(
                c3.knots, c2.knots, err_msg=f"CoilSet {i} Coil {j}"
            )

            coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
            X1 = coords1[:, 0]
            Y1 = coords1[:, 1]
            Z1 = coords1[:, 2]

            coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
            X2 = coords2[:, 0]
            Y2 = coords2[:, 1]
            Z2 = coords2[:, 2]

            coords3 = c3.compute("x", grid=grid, basis="xyz")["x"]
            X3 = coords3[:, 0]
            Y3 = coords3[:, 1]
            Z3 = coords3[:, 2]

            np.testing.assert_allclose(
                c1.current, c2.current, err_msg=f"CoilSet {i} Coil {j}"
            )
            np.testing.assert_allclose(
                c3.current, c2.current, err_msg=f"CoilSet {i} Coil {j}"
            )
            np.testing.assert_allclose(X1, X2, err_msg=f"CoilSet {i} Coil {j}")
            np.testing.assert_allclose(X3, X2, err_msg=f"CoilSet {i} Coil {j}")
            np.testing.assert_allclose(Y1, Y2, err_msg=f"CoilSet {i} Coil {j}")
            np.testing.assert_allclose(Y3, Y2, err_msg=f"CoilSet {i} Coil {j}")
            np.testing.assert_allclose(
                Z1, Z2, atol=2e-7, err_msg=f"CoilSet {i} Coil {j}"
            )
            np.testing.assert_allclose(
                Z3, Z2, atol=2e-7, err_msg=f"CoilSet {i} Coil {j}"
            )

    # check magnetic field from both, check that matches
    grid = LinearGrid(N=200, endpoint=False)
    B1 = coilset.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )
    B2 = coilset2.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )
    B3 = coilset_flat.compute_magnetic_field(
        np.array([[0.7, 0, 0]]), basis="xyz", source_grid=grid
    )

    np.testing.assert_allclose(B1, B2, atol=1e-7)
    np.testing.assert_allclose(B3, B2, atol=1e-7)


@pytest.mark.unit
def test_save_and_load_makegrid_coils_rotated(tmpdir_factory):
    """Test saving and reloading CoilSet linspaced angular from MAKEGRID file."""
    tmpdir = tmpdir_factory.mktemp("coil_files")
    path = tmpdir.join("coils.MAKEGRID_format_angular_coil")

    # make a coilset with angular coilset
    N = 22
    coil = FourierPlanarCoil()
    coil.current = 1
    coilset = CoilSet.linspaced_angular(coil, n=N, angle=2 * np.pi)

    grid = LinearGrid(N=200, endpoint=False)
    coilset.save_in_makegrid_format(str(path), grid=grid, NFP=2)

    coilset2 = CoilSet.from_makegrid_coilfile(str(path))

    # check values at saved points, ensure they match
    for i, (c1, c2) in enumerate(zip(coilset, coilset2)):
        grid = LinearGrid(zeta=coilset2[0].knots, endpoint=False)
        coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
        X1 = coords1[:, 0]
        Y1 = coords1[:, 1]
        Z1 = coords1[:, 2]

        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
        X2 = coords2[:, 0]
        Y2 = coords2[:, 1]
        Z2 = coords2[:, 2]

        np.testing.assert_allclose(c1.current, c2.current, err_msg=f"Coil {i}")
        np.testing.assert_allclose(X1, X2, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Y1, Y2, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Z1, Z2, atol=2e-7, err_msg=f"Coil {i}")

    # check values at interpolated points, ensure they match closely
    grid = LinearGrid(N=51, endpoint=False)
    for c1, c2 in zip(coilset, coilset2):
        coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
        X1 = coords1[:, 0]
        Y1 = coords1[:, 1]
        Z1 = coords1[:, 2]

        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
        X2 = coords2[:, 0]
        Y2 = coords2[:, 1]
        Z2 = coords2[:, 2]

        np.testing.assert_allclose(c1.current, c2.current, err_msg=f"Coil {i}")
        np.testing.assert_allclose(X1, X2, err_msg=f"Coil {i}", atol=1e-16)
        np.testing.assert_allclose(Y1, Y2, err_msg=f"Coil {i}", atol=1e-16)
        np.testing.assert_allclose(Z1, Z2, atol=2e-7, err_msg=f"Coil {i}")

    # check Bnormal on torus and ensure is near zero
    surf = FourierRZToroidalSurface(
        R_lmn=np.array([10, 0.1]),
        Z_lmn=np.array([-0.1]),
        modes_R=np.array([[0, 0], [1, 0]]),
        modes_Z=np.array([[-1, 0]]),
    )

    B_normal, _ = coilset.compute_Bnormal(surf, source_grid=grid)
    np.testing.assert_allclose(B_normal, 0, atol=1e-16)
    B_normal2, _ = coilset2.compute_Bnormal(surf)
    np.testing.assert_allclose(B_normal2, 0, atol=1e-16)

    # check B btwn the two coilsets
    B1 = coilset.compute_magnetic_field(
        np.array([[10, 0, 0]]), basis="xyz", source_grid=32
    )
    B2 = coilset2.compute_magnetic_field(
        np.array([[10, 0, 0]]), basis="xyz", source_grid=1000
    )

    # coilset uses fourier discretization so biot savart is more accurate
    # coilset2 uses hanson hirshman which is only 2nd order
    np.testing.assert_allclose(B1, B2, atol=1e-16, rtol=1e-6)


@pytest.mark.unit
def test_save_and_load_makegrid_coils_rotated_int_grid(tmpdir_factory):
    """Test save/load CoilSet linspaced angular from MAKEGRID file with int grid."""
    tmpdir = tmpdir_factory.mktemp("coil_files")
    path = tmpdir.join("coils.MAKEGRID_format_angular_coil")

    # make a coilset with angular coilset
    N = 10
    coil = FourierPlanarCoil()
    coil.current = 1
    coilset = CoilSet.linspaced_angular(coil, n=N, angle=2 * np.pi)

    grid = 200
    coilset.save_in_makegrid_format(str(path), grid=grid, NFP=2)

    coilset2 = CoilSet.from_makegrid_coilfile(str(path))

    # check values at saved points, ensure they match
    for i, (c1, c2) in enumerate(zip(coilset, coilset2)):
        grid = LinearGrid(zeta=coilset2[0].knots, endpoint=False)
        coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
        X1 = coords1[:, 0]
        Y1 = coords1[:, 1]
        Z1 = coords1[:, 2]

        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
        X2 = coords2[:, 0]
        Y2 = coords2[:, 1]
        Z2 = coords2[:, 2]

        np.testing.assert_allclose(c1.current, c2.current, err_msg=f"Coil {i}")
        np.testing.assert_allclose(X1, X2, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Y1, Y2, err_msg=f"Coil {i}")
        np.testing.assert_allclose(Z1, Z2, atol=2e-7, err_msg=f"Coil {i}")

    # check values at interpolated points, ensure they match closely
    grid = LinearGrid(N=101, endpoint=False)
    for c1, c2 in zip(coilset, coilset2):
        coords1 = c1.compute("x", grid=grid, basis="xyz")["x"]
        X1 = coords1[:, 0]
        Y1 = coords1[:, 1]
        Z1 = coords1[:, 2]

        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]
        X2 = coords2[:, 0]
        Y2 = coords2[:, 1]
        Z2 = coords2[:, 2]

        np.testing.assert_allclose(c1.current, c2.current, err_msg=f"Coil {i}")
        np.testing.assert_allclose(X1, X2, err_msg=f"Coil {i}", atol=1e-16)
        np.testing.assert_allclose(Y1, Y2, err_msg=f"Coil {i}", atol=1e-16)
        np.testing.assert_allclose(Z1, Z2, atol=2e-7, err_msg=f"Coil {i}")

    # check Bnormal on torus and ensure is near zero
    surf = FourierRZToroidalSurface(
        R_lmn=np.array([10, 0.1]),
        Z_lmn=np.array([-0.1]),
        modes_R=np.array([[0, 0], [1, 0]]),
        modes_Z=np.array([[-1, 0]]),
    )

    B_normal, _ = coilset.compute_Bnormal(surf, source_grid=grid)
    np.testing.assert_allclose(B_normal, 0, atol=1e-16)
    B_normal2, _ = coilset2.compute_Bnormal(surf)
    np.testing.assert_allclose(B_normal2, 0, atol=1e-16)

    # check B btwn the two coilsets
    B1 = coilset.compute_magnetic_field(
        np.array([[10, 0, 0]]), basis="xyz", source_grid=grid
    )
    B2 = coilset2.compute_magnetic_field(
        np.array([[10, 0, 0]]), basis="xyz", source_grid=grid
    )

    np.testing.assert_allclose(B1, B2, atol=1e-10)


@pytest.mark.unit
def test_save_makegrid_coils_assert_NFP(tmpdir_factory):
    """Test saving CoilSet that with incompatible NFP throws an error."""
    Ncoils = 22
    input_path = f"./tests/inputs/coils.MAKEGRID_format_{Ncoils}_coils"
    tmpdir = tmpdir_factory.mktemp("coil_files")
    tmp_path = tmpdir.join("coils.MAKEGRID_format_{Ncoils}_coils")
    shutil.copyfile(input_path, tmp_path)

    coilset = CoilSet.from_makegrid_coilfile(str(tmp_path))
    assert len(coilset) == Ncoils  # correct number of coils
    path = tmpdir.join("coils.MAKEGRID_format_desc")
    assert len(coilset) % 3 != 0
    with pytest.raises(AssertionError):
        coilset.save_in_makegrid_format(str(path), NFP=3)


@pytest.mark.unit
def test_load_makegrid_coils_header_asserts(tmpdir_factory):
    """Test loading in CoilSets from incorrect MAKEGRID format files throws error."""
    Ncoils = 22
    input_path = f"./tests/inputs/coils.MAKEGRID_format_{Ncoils}_coils_header_too_long"
    tmpdir = tmpdir_factory.mktemp("coil_files")
    tmp_path = tmpdir.join("coils.MAKEGRID_format_{Ncoils}_coils_header_too_long")
    shutil.copyfile(input_path, tmp_path)
    with pytest.raises(IOError):
        CoilSet.from_makegrid_coilfile(str(tmp_path))
    input_path = f"./tests/inputs/coils.MAKEGRID_format_{Ncoils}_coils_header_too_short"
    shutil.copyfile(input_path, tmp_path)
    with pytest.raises(IOError):
        CoilSet.from_makegrid_coilfile(str(tmp_path))


@pytest.mark.unit
def test_repr():
    """Test string representation of Coil objects."""
    coil = FourierPlanarCoil()
    assert "FourierPlanarCoil" in str(coil)
    assert "current=1" in str(coil)

    coils = CoilSet.linspaced_angular(coil, n=4)
    assert "CoilSet" in str(coils)
    assert "4 submembers" in str(coils)

    coils.name = "MyCoils"
    assert "MyCoils" in str(coils)


@pytest.mark.unit
def test_linking_number():
    """Test calculation of linking number."""
    coil = FourierPlanarCoil(center=[10, 1, 0])
    grid = LinearGrid(N=25)
    # regular modular coilset from symmetry, so that there are 10 coils, half going
    # one way and half going the other way
    coilset = CoilSet(coil, NFP=5, sym=True)
    coil2 = FourierRZCoil()
    # add a coil along the axis that links all the other coils
    coilset2 = MixedCoilSet(coilset, coil2)
    link = coilset2._compute_linking_number(grid=grid)

    # modular coils don't link each other
    np.testing.assert_allclose(link[:-1, :-1], 0, atol=1e-14)
    # axis coil doesn't link itself
    np.testing.assert_allclose(link[-1, -1], 0, atol=1e-14)
    # we expect the axis coil to link all the modular coils, with alternating sign
    # due to alternating orientation of the coils due to symmetry.
    expected = [1, -1] * 5
    np.testing.assert_allclose(link[-1, :-1], expected, rtol=1e-3)



"""Tests for magnetic field classes."""

import matplotlib.pyplot as plt
import numpy as np
import pytest
from diffrax import Dopri5
from scipy.constants import mu_0

from desc.backend import jit, jnp
from desc.basis import DoubleFourierSeries
from desc.compute import rpz2xyz, rpz2xyz_vec, xyz2rpz_vec
from desc.compute.utils import get_params, get_transforms
from desc.derivatives import FiniteDiffDerivative as Derivative
from desc.examples import get
from desc.geometry import FourierRZToroidalSurface, FourierXYZCurve
from desc.grid import LinearGrid
from desc.io import load
from desc.magnetic_fields import (
    CurrentPotentialField,
    DommaschkPotentialField,
    FourierCurrentPotentialField,
    MagneticFieldFromUser,
    OmnigenousField,
    PoloidalMagneticField,
    ScalarPotentialField,
    SplineMagneticField,
    ToroidalMagneticField,
    VectorPotentialField,
    VerticalMagneticField,
    field_line_integrate,
    read_BNORM_file,
    solve_regularized_surface_current,
)
from desc.magnetic_fields._dommaschk import CD_m_k, CN_m_k
from desc.utils import dot


def phi_lm(R, phi, Z, a, m):
    """Scalar potential test function."""
    CNm0 = (R**m - R**-m) / (2 * m)
    Nm1 = CNm0 * Z
    CDm0 = (R**m + R**-m) / 2
    c1 = -m * (m - 1)
    c2 = (m + 1) * (m - 2)
    c3 = m * (m + 1)
    c4 = -(m + 2) * (m - 1)
    CDm1 = (c1 * R ** (m + 2) + c2 * R ** (m) + c3 * R ** (-m + 2) + c4 * R ** (-m)) / (
        8 * m * (m**2 - 1)
    )
    Dm2 = CDm0 * Z**2 / 2 + CDm1
    return phi + a * Dm2 * jnp.sin(m * phi) + a * Nm1 * jnp.cos(m * phi)


a = -1.489
m = 5
args = {"a": a, "m": m}


class TestMagneticFields:
    """Tests for MagneticField classes."""

    @pytest.mark.unit
    def test_basic_fields(self):
        """Tests for basic field types (toroidal, vertical, poloidal)."""
        tfield = ToroidalMagneticField(2, 1)
        vfield = VerticalMagneticField(1)
        pfield = PoloidalMagneticField(2, 1, 2)

        def tfield_A(R, phi, Z, B0=2, R0=1):
            az = -B0 * R0 * jnp.log(R)
            arp = jnp.zeros_like(az)
            A = jnp.array([arp, arp, az]).T
            return A

        tfield_from_A = VectorPotentialField(tfield_A, params={"B0": 2, "R0": 1})

        def vfield_A(R, phi, Z, B0=None):
            coords_rpz = jnp.vstack([R, phi, Z]).T
            coords_xyz = rpz2xyz(coords_rpz)
            ax = B0 / 2 * coords_xyz[:, 1]
            ay = -B0 / 2 * coords_xyz[:, 0]

            az = jnp.zeros_like(ax)
            A = jnp.array([ax, -ay, az]).T
            A = xyz2rpz_vec(A, phi=coords_rpz[:, 1])
            return A

        vfield_params = {"B0": 1}
        vfield_from_A = VectorPotentialField(vfield_A, params=vfield_params)

        np.testing.assert_allclose(tfield([1, 0, 0]), [[0, 2, 0]])
        np.testing.assert_allclose((4 * tfield)([2, 0, 0]), [[0, 4, 0]])
        np.testing.assert_allclose(tfield_from_A([1, 0, 0]), [[0, 2, 0]])
        np.testing.assert_allclose(
            tfield_A(1, 0, 0),
            tfield_from_A.compute_magnetic_vector_potential([1, 0, 0]).squeeze(),
        )
        np.testing.assert_allclose(
            vfield_A(1, 0, 0, **vfield_params),
            vfield_from_A.compute_magnetic_vector_potential([1, 0, 0]),
        )

        np.testing.assert_allclose((tfield + vfield)([1, 0, 0]), [[0, 2, 1]])
        np.testing.assert_allclose(
            (tfield + vfield - pfield)([1, 0, 0.1]), [[0.4, 2, 1]]
        )

    @pytest.mark.unit
    def test_field_from_user(self):
        """Test for MagneticFieldFromUser."""
        tfield = ToroidalMagneticField(2, 1)

        def fun(coords, params):
            R0, B0 = params
            coords = jnp.atleast_2d(jnp.asarray(coords))
            bp = B0 * R0 / coords[:, 0]
            brz = jnp.zeros_like(bp)
            B = jnp.array([brz, bp, brz]).T
            return B

        ufield = MagneticFieldFromUser(fun, [tfield.R0, tfield.B0])
        np.testing.assert_allclose(
            tfield([1, 0, 0]), ufield([1, 0, 0], params=[tfield.R0, tfield.B0])
        )
        np.testing.assert_allclose(
            tfield([1, 1, 0], basis="xyz"),
            ufield([1, 1, 0], params=[tfield.R0, tfield.B0], basis="xyz"),
        )

    @pytest.mark.unit
    def test_combined_fields(self):
        """Tests for sum/scaled fields."""
        tfield = ToroidalMagneticField(2, 1)
        vfield = VerticalMagneticField(1)
        pfield = PoloidalMagneticField(2, 1, 2)
        tfield.R0 = 1
        tfield.B0 = 2
        vfield.B0 = 3.2
        pfield.R0 = 1
        pfield.B0 = 2
        pfield.iota = 1.2
        scaled_field = 3.1 * tfield
        assert scaled_field.B0 == 2
        assert scaled_field.scale == 3.1
        np.testing.assert_allclose(scaled_field([1.0, 0, 0]), np.array([[0, 6.2, 0]]))
        np.testing.assert_allclose(
            scaled_field.compute_magnetic_vector_potential([2.0, 0, 0]),
            np.array([[0, 0, -3.1 * 2 * 1 * np.log(2)]]),
        )

        scaled_field.R0 = 1.3
        scaled_field.scale = 1.0
        np.testing.assert_allclose(scaled_field([1.3, 0, 0]), np.array([[0, 2, 0]]))
        np.testing.assert_allclose(
            scaled_field.compute_magnetic_vector_potential([2.0, 0, 0]),
            np.array([[0, 0, -2 * 1.3 * np.log(2)]]),
        )
        assert scaled_field.optimizable_params == ["B0", "R0", "scale"]
        assert hasattr(scaled_field, "B0")

        sum_field = vfield + pfield + tfield
        sum_field_tv = vfield + tfield  # to test A since pfield does not have A
        assert len(sum_field) == 3
        assert len(sum_field_tv) == 2

        np.testing.assert_allclose(
            sum_field([1.3, 0, 0.0]), [[0.0, 2, 3.2 + 2 * 1.2 * 0.3]]
        )

        tfield_A = np.array([[0, 0, -tfield.B0 * tfield.R0 * np.log(tfield.R0)]])
        x = tfield.R0 * np.cos(np.pi / 4)
        y = tfield.R0 * np.sin(np.pi / 4)
        vfield_A = np.array([[vfield.B0 * y, -vfield.B0 * x, 0]]) / 2

        np.testing.assert_allclose(
            sum_field_tv.compute_magnetic_vector_potential([x, y, 0.0], basis="xyz"),
            tfield_A + vfield_A,
        )

        assert sum_field.optimizable_params == [
            ["B0"],
            ["B0", "R0", "iota"],
            ["B0", "R0"],
        ]
        assert sum_field.dimensions == [
            {"B0": 1},
            {"B0": 1, "R0": 1, "iota": 1},
            {"B0": 1, "R0": 1},
        ]
        assert sum_field.x_idx == [
            {"B0": np.array([0])},
            {"B0": np.array([1]), "R0": np.array([2]), "iota": np.array([3])},
            {"B0": np.array([4]), "R0": np.array([5])},
        ]
        assert sum_field.dim_x == 6
        p = sum_field.pack_params(sum_field.params_dict)
        np.testing.assert_allclose(p, [3.2, 2, 1, 1.2, 2, 1.3])
        p *= 1.3
        sum_field.params_dict = sum_field.unpack_params(p)
        assert sum_field[0].B0 == 1.3 * 3.2
        assert sum_field[1].B0 == 1.3 * 2
        assert sum_field[2].B0 == 1.3 * 2
        sum_field.params_dict = sum_field.unpack_params(p / 1.3)

        del sum_field[-1]
        assert len(sum_field) == 2
        np.testing.assert_allclose(
            sum_field([1.3, 0, 0.0]), [[0.0, 0.0, (3.2 + 2 * 1.2 * 0.3)]]
        )
        assert sum_field.optimizable_params == [
            ["B0"],
            ["B0", "R0", "iota"],
        ]
        sum_field.insert(1, tfield)
        assert len(sum_field) == 3
        np.testing.assert_allclose(
            sum_field([1.3, 0, 0.0]), [[0.0, 2.0, (3.2 + 2 * 1.2 * 0.3)]]
        )
        assert sum_field.optimizable_params == [
            ["B0"],
            ["B0", "R0"],
            ["B0", "R0", "iota"],
        ]
        sum_field[1] = pfield
        sum_field[2] = tfield
        assert sum_field.optimizable_params == [
            ["B0"],
            ["B0", "R0", "iota"],
            ["B0", "R0"],
        ]

    @pytest.mark.unit
    def test_scalar_field(self):
        """Test scalar potential magnetic field against analytic result."""
        field = ScalarPotentialField(phi_lm, args)
        np.testing.assert_allclose(
            field.compute_magnetic_field([1.0, 0, 0]), [[0, 1, 0]]
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field([1.0, np.pi / 4, 0]), [[0, 1, 0]]
        )

    @pytest.mark.unit
    def test_current_potential_field(self):
        """Test current potential magnetic field against analytic result."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        # make a current potential corresponding a purely poloidal current
        G = 10  # net poloidal current
        potential = lambda theta, zeta, G: G * zeta / 2 / jnp.pi
        potential_dtheta = lambda theta, zeta, G: jnp.zeros_like(theta)
        potential_dzeta = lambda theta, zeta, G: G * jnp.ones_like(theta) / 2 / jnp.pi

        params = {"G": -G}
        correct_field = lambda R, phi, Z: jnp.array([[0, mu_0 * G / 2 / jnp.pi / R, 0]])

        field = CurrentPotentialField(
            potential,
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            params=params,
            potential_dtheta=potential_dtheta,
            potential_dzeta=potential_dzeta,
            NFP=surface.NFP,
        )

        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, 0, 0]),
            correct_field(10.0, 0, 0),
            atol=1e-16,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, np.pi / 4, 0]),
            correct_field(10.0, np.pi / 4, 0),
            atol=1e-16,
            rtol=1e-8,
        )

        potential = lambda theta, zeta, G: G * zeta / 2 / jnp.pi * 2
        potential_dtheta = lambda theta, zeta, G: jnp.zeros_like(theta)
        potential_dzeta = (
            lambda theta, zeta, G: G * jnp.ones_like(theta) / 2 / jnp.pi * 2
        )

        correct_field = lambda R, phi, Z: 2 * jnp.array(
            [[0, mu_0 * G / 2 / jnp.pi / R, 0]]
        )
        field.potential = potential
        field.potential_dtheta = potential_dtheta
        field.potential_dzeta = potential_dzeta

        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, 0, 0]),
            correct_field(10.0, 0, 0),
            atol=1e-16,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, np.pi / 4, 0]),
            correct_field(10.0, np.pi / 4, 0),
            atol=1e-16,
            rtol=1e-8,
        )

        field = CurrentPotentialField.from_surface(
            surface=surface,
            potential=potential,
            params=params,
            potential_dtheta=potential_dtheta,
            potential_dzeta=potential_dzeta,
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, 0, 0]),
            correct_field(10.0, 0, 0),
            atol=1e-16,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, np.pi / 4, 0]),
            correct_field(10.0, np.pi / 4, 0),
            atol=1e-16,
            rtol=1e-8,
        )

        # add a ToroidalField and check passing in/not passing in

        B_TF = ToroidalMagneticField(1, 10)
        sumfield = B_TF + field

        np.testing.assert_allclose(
            sumfield.compute_magnetic_field([10.0, 0, 0]),
            correct_field(10.0, 0, 0) + B_TF([10.0, 0, 0]),
            atol=1e-16,
            rtol=1e-8,
        )

        np.testing.assert_allclose(
            sumfield.compute_magnetic_field(
                [10.0, 0, 0],
                source_grid=[None, LinearGrid(M=30, N=30, NFP=surface.NFP)],
            ),
            correct_field(10.0, 0, 0) + B_TF([10.0, 0, 0]),
            atol=1e-16,
            rtol=1e-8,
        )

        with pytest.raises(IOError):
            field.save("test_field.h5")
        with pytest.warns(UserWarning):
            # check that if passing in a longer params dict, that
            # the warning is thrown
            field.params = {"key1": None, "key2": None}
        # check assert callable statement
        with pytest.raises(AssertionError):
            field.potential = 1
        with pytest.raises(AssertionError):
            field.potential_dtheta = 1
        with pytest.raises(AssertionError):
            field.potential_dzeta = 1

    @pytest.mark.unit
    def test_current_potential_vector_potential(self):
        """Test current potential field vector potential against analytic result."""
        R0 = 10
        a = 1
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([R0, a]),
            Z_lmn=jnp.array([0, -a]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        # make a current potential corresponding a purely poloidal current
        G = 100  # net poloidal current
        potential = lambda theta, zeta, G: G * zeta / 2 / jnp.pi
        potential_dtheta = lambda theta, zeta, G: jnp.zeros_like(theta)
        potential_dzeta = lambda theta, zeta, G: G * jnp.ones_like(theta) / 2 / jnp.pi

        params = {"G": -G}

        field = CurrentPotentialField(
            potential,
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            params=params,
            potential_dtheta=potential_dtheta,
            potential_dzeta=potential_dzeta,
            NFP=surface.NFP,
        )
        # test the loop integral of A around a curve encompassing the torus
        # against the analytic result for flux in an ideal toroidal solenoid
        prefactors = mu_0 * G / 2 / jnp.pi
        correct_flux = -2 * np.pi * prefactors * (np.sqrt(R0**2 - a**2) - R0)

        curve = FourierXYZCurve()  # curve to integrate A over
        curve_grid = LinearGrid(zeta=20)
        curve_data = curve.compute(["x", "x_s"], grid=curve_grid, basis="xyz")
        curve_data_rpz = curve.compute(["x", "x_s"], grid=curve_grid, basis="rpz")

        surface_grid = LinearGrid(M=60, N=60, NFP=10)

        A_xyz = field.compute_magnetic_vector_potential(
            curve_data["x"], basis="xyz", source_grid=surface_grid
        )
        A_rpz = field.compute_magnetic_vector_potential(
            curve_data_rpz["x"], basis="rpz", source_grid=surface_grid
        )

        # integrate to get the flux
        flux_xyz = jnp.sum(
            dot(A_xyz, curve_data["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )
        flux_rpz = jnp.sum(
            dot(A_rpz, curve_data_rpz["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )

        np.testing.assert_allclose(correct_flux, flux_xyz, rtol=1e-8)
        np.testing.assert_allclose(correct_flux, flux_rpz, rtol=1e-8)

        field.params["G"] = -2 * field.params["G"]

        A_xyz = field.compute_magnetic_vector_potential(
            curve_data["x"], basis="xyz", source_grid=surface_grid
        )
        A_rpz = field.compute_magnetic_vector_potential(
            curve_data_rpz["x"], basis="rpz", source_grid=surface_grid
        )

        # integrate to get the flux
        flux_xyz = jnp.sum(
            dot(A_xyz, curve_data["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )
        flux_rpz = jnp.sum(
            dot(A_rpz, curve_data_rpz["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )

        np.testing.assert_allclose(-2 * correct_flux, flux_xyz, rtol=1e-8)
        np.testing.assert_allclose(-2 * correct_flux, flux_rpz, rtol=1e-8)

    @pytest.mark.unit
    def test_fourier_current_potential_field(self):
        """Test Fourier current potential magnetic field against analytic result."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        basis = DoubleFourierSeries(M=2, N=2, sym="sin")
        phi_mn = np.ones((basis.num_modes,))
        # make a current potential corresponding a purely poloidal current
        G = 10  # net poloidal current
        correct_field = lambda R, phi, Z: jnp.array([[0, mu_0 * G / 2 / jnp.pi / R, 0]])

        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            I=0,
            G=-G,
            sym_Phi="sin",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        surface_grid = LinearGrid(M=120, N=120, NFP=10)

        phi_mn = np.zeros((basis.num_modes,))

        field.Phi_mn = phi_mn

        field.change_resolution(3, 3)
        field.change_Phi_resolution(2, 2)

        params = get_params(["K", "x"], field)
        transforms = get_transforms(["K", "x"], field, grid=surface_grid)

        np.testing.assert_allclose(
            field.compute_magnetic_field(
                [10.0, 0, 0],
                source_grid=surface_grid,
                params=params,
                transforms=transforms,
            ),
            correct_field(10.0, 0, 0),
            atol=1e-16,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field(
                [10.0, np.pi / 4, 0], source_grid=surface_grid
            ),
            correct_field(10.0, np.pi / 4, 0),
            atol=1e-16,
            rtol=1e-8,
        )

        field.G = -2 * G
        field.I = 0

        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, 0, 0], source_grid=surface_grid),
            correct_field(10.0, 0, 0) * 2,
            atol=1e-16,
            rtol=1e-8,
        )
        # use default grid
        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, np.pi / 4, 0], source_grid=None),
            correct_field(10.0, np.pi / 4, 0) * 2,
            atol=1e-12,
            rtol=1e-8,
        )

        field = FourierCurrentPotentialField.from_surface(
            surface=surface,
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            I=0,
            G=-G,
        )

        np.testing.assert_allclose(
            field.compute_magnetic_field([10.0, 0, 0], source_grid=surface_grid),
            correct_field(10.0, 0, 0),
            atol=1e-16,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            field.compute_magnetic_field(
                [10.0, np.pi / 4, 0], source_grid=surface_grid
            ),
            correct_field(10.0, np.pi / 4, 0),
            atol=1e-16,
            rtol=1e-8,
        )

        K_xyz = field.compute(["K", "x"], basis="xyz", grid=surface_grid)
        K_rpz = field.compute(["K", "x"], basis="rpz", grid=surface_grid)

        np.testing.assert_allclose(
            K_xyz["K"], rpz2xyz_vec(K_rpz["K"], phi=K_rpz["x"][:, 1]), atol=1e-16
        )
        np.testing.assert_allclose(
            K_rpz["K"],
            xyz2rpz_vec(K_xyz["K"], x=K_xyz["x"][:, 0], y=K_xyz["x"][:, 1]),
            atol=1e-16,
        )

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=10)
    def test_fourier_current_potential_field_modular_coil_cut(self):
        """Test Fourier current potential coil cut against analytic solenoid."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([20, 3]),
            Z_lmn=jnp.array([0, -3]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=5,
        )
        # make a current potential corresponding a purely poloidal current
        G = 1e4  # net poloidal current
        correct_field = lambda R, phi, Z: jnp.array([[0, mu_0 * G / 2 / jnp.pi / R, 0]])

        field = FourierCurrentPotentialField(
            I=0,
            G=-G,
            sym_Phi="sin",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=surface.NFP,
        )
        coils = field.to_CoilSet(10, stell_sym=True, show_plots=True).to_FourierXYZ(
            N=2, grid=LinearGrid(N=8), check_intersection=False
        )

        np.testing.assert_allclose(
            coils.compute_magnetic_field(
                [20.0, 0, 0],
            ),
            correct_field(20.0, 0, 0),
            atol=1e-8,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            coils.compute_magnetic_field([20.0, np.pi / 4, 0]),
            correct_field(20.0, np.pi / 4, 0),
            atol=1e-8,
            rtol=1e-8,
        )
        return plt.gcf()

    @pytest.mark.unit
    def test_fourier_current_potential_field_helical_coil_cut(self):
        """Test Fourier current potential helix coil cut against analytic solenoid."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([20, 3]),
            Z_lmn=jnp.array([0, -3]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        # make a current potential corresponding a helical current
        # with a sharp pitch (approximating a toroidal solenoid)
        G = 2e2  # net poloidal current
        I = 1
        correct_field = lambda R, phi, Z: jnp.array([[0, mu_0 * G / 2 / jnp.pi / R, 0]])

        field = FourierCurrentPotentialField(
            I=I,
            G=-G,
            sym_Phi="sin",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )

        coils = field.to_CoilSet(1)

        np.testing.assert_allclose(
            coils.compute_magnetic_field([20.0, 0, 0], source_grid=LinearGrid(N=700)),
            correct_field(20.0, 0, 0),
            atol=2e-8,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            coils.compute_magnetic_field(
                [20.0, np.pi / 4, 0], source_grid=LinearGrid(N=700)
            ),
            correct_field(20.0, np.pi / 4, 0),
            atol=2e-8,
            rtol=1e-8,
        )
        # check with opposite helicity current
        field = FourierCurrentPotentialField(
            I=I,
            G=G,
            sym_Phi="sin",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )

        coils = field.to_CoilSet(1)

        np.testing.assert_allclose(
            -coils.compute_magnetic_field([20.0, 0, 0], source_grid=LinearGrid(N=700)),
            correct_field(20.0, 0, 0),
            atol=2e-8,
            rtol=1e-8,
        )
        np.testing.assert_allclose(
            -coils.compute_magnetic_field(
                [20.0, np.pi / 4, 0], source_grid=LinearGrid(N=700)
            ),
            correct_field(20.0, np.pi / 4, 0),
            atol=2e-8,
            rtol=1e-8,
        )

    @pytest.mark.unit
    def test_fourier_current_potential_vector_potential(self):
        """Test Fourier current potential vector potential against analytic result."""
        R0 = 10
        a = 1
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([R0, a]),
            Z_lmn=jnp.array([0, -a]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )

        basis = DoubleFourierSeries(M=2, N=2, sym="sin")
        phi_mn = np.ones((basis.num_modes,))
        # make a current potential corresponding a purely poloidal current
        G = 100  # net poloidal current

        # test the loop integral of A around a curve encompassing the torus
        # against the analytic result for flux in an ideal toroidal solenoid
        ## expression for flux inside of toroidal solenoid of radius a
        prefactors = mu_0 * G / 2 / jnp.pi
        correct_flux = -2 * np.pi * prefactors * (np.sqrt(R0**2 - a**2) - R0)

        curve = FourierXYZCurve()  # curve to integrate A over
        curve_grid = LinearGrid(zeta=20)
        curve_data = curve.compute(["x", "x_s"], grid=curve_grid)
        curve_data_rpz = curve.compute(["x", "x_s"], grid=curve_grid, basis="rpz")

        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            I=0,
            G=-G,  # to get a positive B_phi, we must put G negative
            # since -G is the net poloidal current on the surface
            # ( with  G=-(net_current) meaning that we have net_current
            # flowing poloidally (in clockwise direction) around torus)
            sym_Phi="sin",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        surface_grid = LinearGrid(M=60, N=60, NFP=10)

        phi_mn = np.zeros((basis.num_modes,))

        field.Phi_mn = phi_mn

        field.change_resolution(3, 3)
        field.change_Phi_resolution(2, 2)

        A_xyz = field.compute_magnetic_vector_potential(
            curve_data["x"], basis="xyz", source_grid=surface_grid
        )
        A_rpz = field.compute_magnetic_vector_potential(
            curve_data_rpz["x"], basis="rpz", source_grid=surface_grid
        )

        # integrate to get the flux
        flux_xyz = jnp.sum(
            dot(A_xyz, curve_data["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )
        flux_rpz = jnp.sum(
            dot(A_rpz, curve_data_rpz["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )

        np.testing.assert_allclose(correct_flux, flux_xyz, rtol=1e-8)
        np.testing.assert_allclose(correct_flux, flux_rpz, rtol=1e-8)

        field.G = -2 * field.G
        field.I = 0

        A_xyz = field.compute_magnetic_vector_potential(
            curve_data["x"], basis="xyz", source_grid=surface_grid
        )
        A_rpz = field.compute_magnetic_vector_potential(
            curve_data_rpz["x"], basis="rpz", source_grid=surface_grid
        )

        # integrate to get the flux
        flux_xyz = jnp.sum(
            dot(A_xyz, curve_data["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )
        flux_rpz = jnp.sum(
            dot(A_rpz, curve_data_rpz["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )

        np.testing.assert_allclose(-2 * correct_flux, flux_xyz, rtol=1e-8)
        np.testing.assert_allclose(-2 * correct_flux, flux_rpz, rtol=1e-8)

        field = FourierCurrentPotentialField.from_surface(
            surface=surface,
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            I=0,
            G=-G,
        )

        A_xyz = field.compute_magnetic_vector_potential(
            curve_data["x"], basis="xyz", source_grid=surface_grid
        )
        A_rpz = field.compute_magnetic_vector_potential(
            curve_data_rpz["x"], basis="rpz", source_grid=surface_grid
        )

        # integrate to get the flux
        flux_xyz = jnp.sum(
            dot(A_xyz, curve_data["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )
        flux_rpz = jnp.sum(
            dot(A_rpz, curve_data_rpz["x_s"], axis=-1) * curve_grid.spacing[:, 2]
        )

        np.testing.assert_allclose(correct_flux, flux_xyz, rtol=1e-8)
        np.testing.assert_allclose(correct_flux, flux_rpz, rtol=1e-8)

    @pytest.mark.unit
    def test_fourier_current_potential_field_symmetry(self):
        """Test Fourier current potential magnetic field Phi symmetry logic."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        basis = DoubleFourierSeries(M=2, N=2, sym="cos")
        phi_mn = np.ones((basis.num_modes,))
        # make a current potential corresponding a purely poloidal current
        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            sym_Phi="cos",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        assert field.sym_Phi == "cos"

        basis = DoubleFourierSeries(M=2, N=2, sym=False)
        phi_mn = np.ones((basis.num_modes,))
        # make a current potential corresponding a purely poloidal current
        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        assert field.sym_Phi is False

        # check error thrown if new array is different size than old
        with pytest.raises(ValueError):
            field.Phi_mn = np.ones((basis.num_modes + 1,))

    @pytest.mark.unit
    def test_io_fourier_current_field(self, tmpdir_factory):
        """Test that i/o works for FourierCurrentPotentialField."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        basis = DoubleFourierSeries(M=2, N=2, sym="cos")
        phi_mn = np.ones((basis.num_modes,))
        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            G=1000,
            I=-50,
            modes_Phi=basis.modes[:, 1:],
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        tmpdir = tmpdir_factory.mktemp("test_io_fourier_current_field")
        field.save(tmpdir.join("test_field.h5"))
        field2 = load(tmpdir.join("test_field.h5"))
        assert field.equiv(field2)

    @pytest.mark.unit
    def test_fourier_current_potential_field_asserts(self):
        """Test Fourier current potential magnetic field assert statements."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        basis = DoubleFourierSeries(M=2, N=2, sym="sin")
        phi_mn = np.ones((basis.num_modes,))
        # make a current potential corresponding a purely poloidal current
        G = 10  # net poloidal current

        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            I=0,
            G=-G,
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        # check that we can change I,G correctly

        # with scalars
        field.I = 1
        field.G = 2
        np.testing.assert_allclose(field.I, 1)
        np.testing.assert_allclose(field.G, 2)
        # with 0D array
        field.I = np.array(1)
        field.G = np.array(2)
        np.testing.assert_allclose(field.I, 1)
        np.testing.assert_allclose(field.G, 2)
        # with 1D array of size 1
        field.I = np.array([1])
        field.G = np.array([2])
        np.testing.assert_allclose(field.I, 1)
        np.testing.assert_allclose(field.G, 2)

        # check that we can't set it with a size>1 array
        with pytest.raises(TypeError):
            field.I = np.array([1, 2])
        with pytest.raises(TypeError):
            field.G = np.array([1, 2])

        # check that we cant initialize with different size
        # Phi_mn and Phi_modes arrays
        with pytest.raises(AssertionError):
            field = FourierCurrentPotentialField(
                Phi_mn=phi_mn[0:-1],  # too short by 1
                modes_Phi=basis.modes[:, 1:],
                I=0,
                G=-G,
                R_lmn=surface.R_lmn,
                Z_lmn=surface.Z_lmn,
                modes_R=surface._R_basis.modes[:, 1:],
                modes_Z=surface._Z_basis.modes[:, 1:],
                NFP=10,
            )

    @pytest.mark.unit
    def test_change_Phi_basis_fourier_current_field(self):
        """Test that change_Phi_resolution works for FourierCurrentPotentialField."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        M = N = 2
        basis = DoubleFourierSeries(M=M, N=N, NFP=surface.NFP, sym="cos")

        phi_mn = np.ones((basis.num_modes,))
        field = FourierCurrentPotentialField(
            Phi_mn=phi_mn,
            modes_Phi=basis.modes[:, 1:],
            sym_Phi="cos",
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )

        np.testing.assert_allclose(abs(np.max(field.Phi_basis.modes[:, 1:])), M)
        np.testing.assert_allclose(field.Phi_basis.modes, basis.modes)
        assert field.Phi_basis.sym == "cos"

        M = N = 5
        basis = DoubleFourierSeries(M=M, N=N, NFP=surface.NFP, sym="cos")
        field.change_Phi_resolution(M=M, N=N)

        np.testing.assert_allclose(abs(np.max(field.Phi_basis.modes[:, 1:])), M)
        np.testing.assert_allclose(field.Phi_basis.modes, basis.modes)
        assert field.Phi_basis.sym == "cos"

        M = 7
        N = 9
        basis = DoubleFourierSeries(M=M, N=N, NFP=surface.NFP, sym="cos")
        field.change_Phi_resolution(M=M, N=N)

        np.testing.assert_allclose(abs(np.max(field.Phi_basis.modes[:, 1])), M)
        np.testing.assert_allclose(abs(np.max(field.Phi_basis.modes[:, 2])), N)
        np.testing.assert_allclose(field.Phi_basis.modes, basis.modes)
        assert field.Phi_basis.sym == "cos"

        M = 3
        N = 3
        basis = DoubleFourierSeries(M=M, N=N, NFP=surface.NFP, sym="sin")
        field.change_Phi_resolution(M=M, N=N, sym_Phi="sin")

        np.testing.assert_allclose(abs(np.max(field.Phi_basis.modes[:, 1])), M)
        np.testing.assert_allclose(abs(np.max(field.Phi_basis.modes[:, 2])), N)
        np.testing.assert_allclose(field.Phi_basis.modes, basis.modes)
        assert field.Phi_basis.sym == "sin"

    @pytest.mark.unit
    def test_init_Phi_mn_fourier_current_field(self):
        """Test initial Phi_mn size is correct for FourierCurrentPotentialField."""
        surface = FourierRZToroidalSurface(
            R_lmn=jnp.array([10, 1]),
            Z_lmn=jnp.array([0, -1]),
            modes_R=jnp.array([[0, 0], [1, 0]]),
            modes_Z=jnp.array([[0, 0], [-1, 0]]),
            NFP=10,
        )
        init_modes = np.array([[1, 1], [3, 3]])
        init_coeffs = np.array([1, 1])

        field = FourierCurrentPotentialField(
            Phi_mn=init_coeffs,
            modes_Phi=init_modes,
            R_lmn=surface.R_lmn,
            Z_lmn=surface.Z_lmn,
            modes_R=surface._R_basis.modes[:, 1:],
            modes_Z=surface._Z_basis.modes[:, 1:],
            NFP=10,
        )
        assert field.Phi_mn.size == field.Phi_basis.num_modes
        inds_nonzero = []
        for coef, modes in zip(init_coeffs, init_modes):
            ind = field.Phi_basis.get_idx(M=modes[0], N=modes[1])
            assert coef == field.Phi_mn[ind]
            inds_nonzero.append(ind)
        inds_zero = np.setdiff1d(np.arange(field.Phi_basis.num_modes), inds_nonzero)
        np.testing.assert_allclose(field.Phi_mn[inds_zero], 0)
        # ensure can compute field at a point without incompatible size error
        field.compute_magnetic_field([10.0, 0, 0])

    @pytest.mark.unit
    def test_fourier_current_potential_field_coil_cut_warnings(self):
        """Test Fourier current potential coil cut method warning."""
        curr = 1e4
        # with this choice of Phi_mn, the constant Phi contours
        # move so much that they intersect the boundaries of where we
        # plot them, that should return a warning
        # TODO: When we switch from the visual coil cutting method, remove this
        field = FourierCurrentPotentialField(
            I=curr,
            G=curr,
            Phi_mn=np.array([-4 * curr / 13]),
            modes_Phi=np.array([[-1, 0]]),
        )

        with pytest.warns(
            UserWarning,
            match="Detected",
        ):
            field.to_CoilSet(2)

    @pytest.mark.slow
    @pytest.mark.unit
    def test_spline_field(self, tmpdir_factory):
        """Test accuracy of spline magnetic field."""
        field1 = ScalarPotentialField(phi_lm, args, NFP=5)
        R = np.linspace(0.5, 1.5, 20)
        Z = np.linspace(-1.5, 1.5, 20)
        p = np.linspace(0, 2 * np.pi / 5, 40)
        field2 = SplineMagneticField.from_field(field1, R, p, Z)

        np.testing.assert_allclose(
            field1([1.0, 1.0, 1.0]), field2([1.0, 1.0, 1.0]), rtol=1e-2, atol=1e-2
        )

        extcur = [4700.0, 1000.0]
        mgrid = "tests/inputs/mgrid_test.nc"
        field3 = SplineMagneticField.from_mgrid(mgrid, extcur)
        # test saving and loading from mgrid
        tmpdir = tmpdir_factory.mktemp("spline_mgrid_with_A")
        path = tmpdir.join("spline_mgrid_with_A.nc")
        field3.save_mgrid(
            path,
            Rmin=np.min(field3._R),
            Rmax=np.max(field3._R),
            Zmin=np.min(field3._Z),
            Zmax=np.max(field3._Z),
            nR=field3._R.size,
            nZ=field3._Z.size,
            nphi=field3._phi.size,
        )
        # no need for extcur b/c is saved in "raw" format, no need to scale again
        field4 = SplineMagneticField.from_mgrid(path)
        attrs_4d = ["_AR", "_Aphi", "_AZ", "_BR", "_Bphi", "_BZ"]
        for attr in attrs_4d:
            np.testing.assert_allclose(
                (getattr(field3, attr) * np.array(extcur)).sum(axis=-1),
                getattr(field4, attr).squeeze(),
                err_msg=attr,
            )
        attrs_3d = ["_R", "_phi", "_Z"]
        for attr in attrs_3d:
            np.testing.assert_allclose(getattr(field3, attr), getattr(field4, attr))

        r = 0.70
        p = 0
        z = 0
        # use finite diff derivatives to check A accuracy
        tfield_A = lambda R, phi, Z: field3.compute_magnetic_vector_potential(
            jnp.vstack([R, phi, Z]).T
        )
        funR = lambda x: tfield_A(x, p, z)
        funP = lambda x: tfield_A(r, x, z)
        funZ = lambda x: tfield_A(r, p, x)

        ap = tfield_A(r, p, z)[:, 1]

        # these are the gradients of each component of A
        dAdr = Derivative.compute_jvp(funR, 0, (jnp.ones_like(r),), r)
        dAdp = Derivative.compute_jvp(funP, 0, (jnp.ones_like(p),), p)
        dAdz = Derivative.compute_jvp(funZ, 0, (jnp.ones_like(z),), z)

        # form the B components with the appropriate combinations
        B2 = jnp.array(
            [
                dAdp[:, 2] / r - dAdz[:, 1],
                dAdz[:, 0] - dAdr[:, 2],
                dAdr[:, 1] + (ap - dAdp[:, 0]) / r,
            ]
        ).T

        np.testing.assert_allclose(
            field3([0.70, 0, 0]), np.array([[0, -0.671, 0.0858]]), rtol=1e-3, atol=1e-8
        )

        np.testing.assert_allclose(field3([0.70, 0, 0]), B2, rtol=1e-3, atol=5e-3)

        field3.currents *= 2
        np.testing.assert_allclose(
            field3([0.70, 0, 0]),
            2 * np.array([[0, -0.671, 0.0858]]),
            rtol=1e-3,
            atol=1e-8,
        )

    @pytest.mark.unit
    def test_spline_field_axisym(self):
        """Test computing axisymmetric magnetic field using SplineMagneticField."""
        extcur = [
            -1.370985e03,
            -1.609154e03,
            -2.751331e03,
            -2.524384e03,
            -3.435372e03,
            -3.466123e03,
            3.670919e03,
            3.450196e03,
            2.908027e03,
            3.404695e03,
            -4.148967e03,
            -4.294406e03,
            -3.059939e03,
            -2.990609e03,
            3.903818e03,
            3.727301e03,
            -3.049484e03,
            -3.086940e03,
            -1.488703e07,
            -2.430716e04,
            -2.380229e04,
        ]
        with pytest.warns(UserWarning):
            # user warning because saved mgrid no vector potential
            field = SplineMagneticField.from_mgrid(
                "tests/inputs/mgrid_d3d.nc", extcur=extcur
            )
        # make sure field is invariant to shift in phi
        B1 = field.compute_magnetic_field(np.array([1.75, 0.0, 0.0]))
        B2 = field.compute_magnetic_field(np.array([1.75, 1.0, 0.0]))
        np.testing.assert_allclose(B1, B2)

        # test the error when no vec pot values exist
        with pytest.raises(ValueError, match="no vector potential"):
            field.compute_magnetic_vector_potential(np.array([1.75, 0.0, 0.0]))

    @pytest.mark.unit
    def test_field_line_integrate(self):
        """Test field line integration."""
        # q=4, field line should rotate 1/4 turn after 1 toroidal transit
        # from outboard midplane to top center
        field = ToroidalMagneticField(2, 10) + PoloidalMagneticField(2, 10, 0.25)
        r0 = [10.001]
        z0 = [0.0]
        phis = [0, 2 * np.pi]
        r, z = field_line_integrate(r0, z0, phis, field)
        np.testing.assert_allclose(r[-1], 10, rtol=1e-6, atol=1e-6)
        np.testing.assert_allclose(z[-1], 0.001, rtol=1e-6, atol=1e-6)

    @pytest.mark.unit
    def test_field_line_integrate_long(self):
        """Test field line integration for long distance along line."""
        # q=4, field line should rotate 1/4 turn after 1 toroidal transit
        # from outboard midplane to top center
        field = ToroidalMagneticField(2, 10) + PoloidalMagneticField(2, 10, 0.25)
        r0 = [10.001]
        z0 = [0.0]
        phis = [0, 2 * np.pi * 25]
        r, z = field_line_integrate(r0, z0, phis, field, solver=Dopri5())
        np.testing.assert_allclose(r[-1], 10, rtol=1e-6, atol=1e-6)
        np.testing.assert_allclose(z[-1], 0.001, rtol=1e-6, atol=1e-6)

    @pytest.mark.unit
    def test_field_line_integrate_early_terminate_default(self):
        """Test field line integration with default early termination criterion."""
        # q=4, field line should rotate 1/4 turn after 1 toroidal transit
        # from outboard midplane to top center
        # early terminate when it crosses towards the inboard side (R=10),
        field1 = ToroidalMagneticField(2, 10) + PoloidalMagneticField(2, 10, 0.25)
        # make a SplineMagneticField only defined in a tiny region around initial point
        field = SplineMagneticField.from_field(
            field=field1,
            R=np.linspace(10.0, 10.005, 40),
            phi=np.linspace(0, 2 * np.pi, 40),
            Z=np.linspace(-5e-3, 5e-3, 40),
            extrap=True,
        )
        r0 = [10.001]
        z0 = [0.0]
        phis = [0, 2 * np.pi, 2 * np.pi * 2]

        r, z = field_line_integrate(
            r0,
            z0,
            phis,
            field,
            bounds_R=(np.min(field._R), np.max(field._R)),
            bounds_Z=(np.min(field._Z), np.max(field._Z)),
            min_step_size=1e-2,
        )
        np.testing.assert_allclose(r[1], 10, rtol=1e-6, atol=1e-6)
        np.testing.assert_allclose(z[1], 0.001, rtol=1e-6, atol=1e-6)
        # if early terinated, the values at the un-integrated phi points are inf
        assert np.isnan(r[-1])
        assert np.isnan(z[-1])

    @pytest.mark.unit
    def test_Bnormal_calculation(self):
        """Tests Bnormal calculation for simple toroidal field."""
        tfield = ToroidalMagneticField(2, 1)
        surface = get("DSHAPE").surface
        Bnorm, _ = tfield.compute_Bnormal(surface)
        # should have 0 Bnormal because surface is axisymmetric
        np.testing.assert_allclose(Bnorm, 0, atol=1e-14)

    @pytest.mark.unit
    def test_Bnormal_save_and_load_DSHAPE(self, tmpdir_factory):
        """Tests Bnormal save/load for simple toroidal field with DSHAPE."""
        ### test on simple field first with tokamak
        tmpdir = tmpdir_factory.mktemp("BNORM_files")
        path = tmpdir.join("BNORM_desc.txt")
        tfield = ToroidalMagneticField(2, 1)
        eq = get("DSHAPE")
        grid = LinearGrid(rho=np.array(1.0), M=20, N=20, NFP=eq.NFP)
        x = eq.surface.compute("x", grid=grid, basis="rpz")["x"]
        Bnorm, x_from_Bnorm = tfield.compute_Bnormal(
            eq.surface, eval_grid=grid, source_grid=grid, basis="rpz"
        )

        # make sure x calculation is the same
        np.testing.assert_allclose(x[:, 0], x_from_Bnorm[:, 0], atol=1e-16)
        np.testing.assert_allclose(x[:, 2], x_from_Bnorm[:, 2], atol=1e-16)

        np.testing.assert_allclose(
            x[:, 1] % (2 * np.pi), x_from_Bnorm[:, 1] % (2 * np.pi), atol=1e-16
        )

        # should have 0 Bnormal because surface is axisymmetric
        np.testing.assert_allclose(Bnorm, 0, atol=1e-14)

        tfield.save_BNORM_file(eq, path, scale_by_curpol=False)
        Bnorm_from_file = read_BNORM_file(path, eq, grid, scale_by_curpol=False)
        np.testing.assert_allclose(Bnorm, Bnorm_from_file, atol=1e-14)

        # check that loading/saving with scale_by_curpol true
        # but no eq passed raises error
        with pytest.raises(RuntimeError):
            Bnorm_from_file = read_BNORM_file(path, eq.surface, grid)
        with pytest.raises(RuntimeError):
            tfield.save_BNORM_file(eq.surface, path)

    @pytest.mark.unit
    def test_Bnormal_save_and_load_HELIOTRON(self, tmpdir_factory):
        """Tests Bnormal save/load for simple toroidal field with HELIOTRON."""
        # test on simple field with stellarator
        tmpdir = tmpdir_factory.mktemp("BNORM_files")
        path = tmpdir.join("BNORM_desc_heliotron.txt")
        tfield = ToroidalMagneticField(2, 1)
        eq = get("HELIOTRON")
        grid = LinearGrid(rho=np.array(1.0), M=20, N=20, NFP=eq.NFP)
        x = eq.surface.compute("x", grid=grid, basis="xyz")["x"]
        Bnorm, x_from_Bnorm = tfield.compute_Bnormal(
            eq.surface, eval_grid=grid, basis="xyz"
        )

        # make sure x calculation is the same
        np.testing.assert_allclose(x, x_from_Bnorm, atol=1e-16)

        tfield.save_BNORM_file(eq, path, 40, 40)
        Bnorm_from_file = read_BNORM_file(path, eq, grid)
        np.testing.assert_allclose(Bnorm, Bnorm_from_file, atol=1e-8)

        asym_surf = FourierRZToroidalSurface(sym=False)
        with pytest.raises(AssertionError, match="sym"):
            Bnorm_from_file = read_BNORM_file(path, asym_surf, grid)

    @pytest.mark.unit
    def test_spline_field_jit(self):
        """Test that the spline field can be passed to a jitted function."""
        extcur = [4700.0, 1000.0]
        mgrid = "tests/inputs/mgrid_test.nc"
        field = SplineMagneticField.from_mgrid(mgrid, extcur)

        x = jnp.array([0.70, 0, 0])

        @jit
        def foo(field, x):
            return field.compute_magnetic_field(x)

        np.testing.assert_allclose(
            foo(field, x), np.array([[0, -0.671, 0.0858]]), rtol=1e-3, atol=1e-8
        )

    @pytest.mark.unit
    def test_mgrid_io(self, tmpdir_factory):
        """Test saving to and loading from an mgrid file."""
        tmpdir = tmpdir_factory.mktemp("mgrid_dir")
        path = tmpdir.join("mgrid.nc")

        # field to test on
        toroidal_field = ToroidalMagneticField(B0=1, R0=5)
        poloidal_field = PoloidalMagneticField(B0=1, R0=5, iota=2 / np.pi)
        vertical_field = VerticalMagneticField(B0=0.2)
        save_field = toroidal_field + poloidal_field + vertical_field

        # save and load mgrid file
        Rmin = 3
        Rmax = 7
        Zmin = -2
        Zmax = 2
        with pytest.raises(NotImplementedError):
            # Raises error because poloidal field has no vector potential
            # and so cannot save the vector potential
            save_field.save_mgrid(path, Rmin, Rmax, Zmin, Zmax)
        save_field.save_mgrid(path, Rmin, Rmax, Zmin, Zmax, save_vector_potential=False)
        with pytest.warns(UserWarning):
            # user warning because saved mgrid has no vector potential
            # and so cannot load the vector potential
            load_field = SplineMagneticField.from_mgrid(path)

        # check that the fields are the same
        num_nodes = 50
        grid = np.array(
            [
                np.linspace(Rmin, Rmax, num_nodes),
                np.linspace(0, 2 * np.pi, num_nodes, endpoint=False),
                np.linspace(Zmin, Zmax, num_nodes),
            ]
        ).T
        B_saved = save_field.compute_magnetic_field(grid)
        B_loaded = load_field.compute_magnetic_field(grid)
        np.testing.assert_allclose(B_loaded, B_saved, rtol=1e-6)

    @pytest.mark.unit
    def test_omnigenous_field_change_resolution_B(self):
        """Test OmnigenousField.change_resolution() of the B_lm parameters."""
        L_B_old = 1
        L_B_new = 2
        M_B_old = 3
        M_B_new = 6
        NFP = 4
        field = OmnigenousField(
            L_B=L_B_old,
            M_B=M_B_old,
            L_x=0,
            M_x=0,
            N_x=0,
            NFP=NFP,
            helicity=(0, NFP),
            B_lm=np.array([0.9, 1.0, 1.1, 0.2, 0.05, -0.2]),
        )
        grid_axis = LinearGrid(rho=[0.0], M=50)
        grid_half = LinearGrid(rho=[0.5], M=50)
        grid_lcfs = LinearGrid(rho=[1.0], M=50)
        B_axis_lowres = field.compute("|B|", grid=grid_axis)["|B|"]
        B_half_lowres = field.compute("|B|", grid=grid_half)["|B|"]
        B_lcfs_lowres = field.compute("|B|", grid=grid_lcfs)["|B|"]
        field.change_resolution(L_B=L_B_new, M_B=M_B_new)
        B_axis_highres = field.compute("|B|", grid=grid_axis)["|B|"]
        B_half_highres = field.compute("|B|", grid=grid_half)["|B|"]
        B_lcfs_highres = field.compute("|B|", grid=grid_lcfs)["|B|"]
        np.testing.assert_allclose(B_axis_lowres, B_axis_highres, rtol=6e-3)
        np.testing.assert_allclose(B_half_lowres, B_half_highres, rtol=3e-3)
        np.testing.assert_allclose(B_lcfs_lowres, B_lcfs_highres, rtol=4e-3)

    @pytest.mark.unit
    def test_solve_current_potential_warnings_and_errors(self):
        """Test solve current potential warnings/errors."""
        field = FourierCurrentPotentialField(I=0, G=1, sym_Phi="sin")
        eq = get("SOLOVEV")
        with pytest.warns(UserWarning, match="Reducing radial"):
            eq.change_resolution(L=1, M=1, N=1, L_grid=1, M_grid=1, N_grid=1)
        with pytest.raises(ValueError, match="length"):
            solve_regularized_surface_current(field, eq, current_helicity=(1, 1, 1))
        with pytest.raises(ValueError, match="integer"):
            solve_regularized_surface_current(field, eq, current_helicity=(1.2, 1))
        with pytest.raises(ValueError, match="regularization"):
            solve_regularized_surface_current(
                field, eq, regularization_type="not a valid option"
            )
        with pytest.raises(ValueError, match="Expected Fourier"):
            solve_regularized_surface_current(ToroidalMagneticField(1, 1), eq)
        with pytest.raises(AssertionError, match="Expected MagneticField"):
            solve_regularized_surface_current(field, eq, external_field=eq)
        field = FourierCurrentPotentialField(I=0, G=1, sym_Phi="cos")
        grid = LinearGrid(M=1, N=1)
        # nested with pytest.warns, if a warning is not detected it is
        # re-emitted and goes through the higher level context,
        #  this lets us test 3 different warnings with one fxn
        # call here
        with pytest.warns(UserWarning, match="Detected"):
            with pytest.warns(UserWarning, match="Pressure"):
                with pytest.warns(UserWarning, match="Current"):
                    solve_regularized_surface_current(
                        field,
                        eq,
                        eval_grid=grid,
                        source_grid=grid,
                        vc_source_grid=grid,
                        verbose=0,
                        vacuum=True,
                    )


@pytest.mark.unit
def test_dommaschk_CN_CD_m_0():
    """Test of CD_m_k and CN_m_k when k=0."""
    # based off eqn 8 and 9 of Dommaschk paper
    # https://doi.org/10.1016/0010-4655(86)90109-8
    for m in range(1, 6):
        # test of CD_m_k based off eqn 8
        R = np.linspace(0.1, 1, 100)
        res1 = CD_m_k(R, m, 0)
        res2 = 0.5 * (R**m + R ** (-m))
        np.testing.assert_allclose(res1, res2)

        # test of CN_m_k based off eqn 9
        res1 = CN_m_k(R, m, 0)
        res2 = 0.5 * (R**m - R ** (-m)) / m
        np.testing.assert_allclose(res1, res2, atol=1e-15)


@pytest.mark.unit
def test_dommaschk_field_errors():
    """Test the assert statements of the DommaschkField function."""
    ms = [1]
    ls = [1]
    a_arr = [1]
    b_arr = [1]
    c_arr = [1]
    d_arr = [1, 1]  # length is not equal to the rest
    with pytest.raises(AssertionError, match="size"):
        DommaschkPotentialField(ms, ls, a_arr, b_arr, c_arr, d_arr)
    d_arr = [1]  # test with incorrect NFP
    with pytest.raises(AssertionError, match="desired NFP"):
        DommaschkPotentialField(ms, ls, a_arr, b_arr, c_arr, d_arr, NFP=2)
    ms = [-1]  # negative mode number
    with pytest.raises(AssertionError, match=">= 0"):
        DommaschkPotentialField(ms, ls, a_arr, b_arr, c_arr, d_arr)


@pytest.mark.unit
def test_dommaschk_radial_field():
    """Test the Dommaschk potential for a pure toroidal (Bphi~1/R) field."""
    phi = np.linspace(0, 2 * np.pi, 10)
    R = np.linspace(0.1, 1.5, 50)
    Z = np.linspace(-0.05, 0.5, 50)
    R, phi, Z = np.meshgrid(R, phi, Z)
    coords = np.vstack((R.flatten(), phi.flatten(), Z.flatten())).T

    ms = [0]
    ls = [0]
    a_arr = [0]
    b_arr = [0]
    c_arr = [0]
    d_arr = [0]
    B = DommaschkPotentialField(ms, ls, a_arr, b_arr, c_arr, d_arr)
    B_dom = B.compute_magnetic_field(coords)
    np.testing.assert_allclose(B_dom[:, 0], 0)
    np.testing.assert_array_equal(B_dom[:, 1], 1 / R.flatten())
    np.testing.assert_allclose(B_dom[:, 2], 0)


@pytest.mark.unit
def test_dommaschk_vertical_field():
    """Test the Dommaschk potential for a 1/R toroidal + pure vertical field."""
    phi = jnp.linspace(0, 2 * jnp.pi, 10)
    R = jnp.linspace(0.1, 1.5, 50)
    Z = jnp.linspace(-0.5, 0.5, 50)
    R, phi, Z = jnp.meshgrid(R, phi, Z)
    coords = jnp.vstack((R.flatten(), phi.flatten(), Z.flatten())).T

    ms = [0]
    ls = [1]
    a_arr = [1]
    b_arr = [0]
    c_arr = [0]
    d_arr = [0]
    B = DommaschkPotentialField(ms, ls, a_arr, b_arr, c_arr, d_arr)
    B_dom = B.compute_magnetic_field(coords)
    ones = jnp.ones_like(B_dom[:, 0])
    np.testing.assert_allclose(B_dom[:, 0], 0, atol=1e-14)
    np.testing.assert_allclose(B_dom[:, 1], 1 / R.flatten(), atol=1e-14)
    np.testing.assert_allclose(B_dom[:, 2], ones, atol=5e-15)


@pytest.mark.unit
def test_dommaschk_fit_vertical_and_toroidal_field():
    """Test the Dommaschk potential fit for a toroidal and a vertical field."""
    phi = np.linspace(0, 2 * np.pi, 3)
    R = np.linspace(0.1, 1.5, 3)
    Z = np.linspace(-0.5, 0.5, 3)
    R, phi, Z = np.meshgrid(R, phi, Z)
    coords = np.vstack((R.flatten(), phi.flatten(), Z.flatten())).T

    max_l = 1
    max_m = 1
    B0 = 2  # scale strength for to 1/R field to fit
    B0_Z = 1  # scale strength for to uniform vertical field to fit
    field = ToroidalMagneticField(B0=B0, R0=1) + VerticalMagneticField(B0=B0_Z)

    B = DommaschkPotentialField.fit_magnetic_field(
        field, coords, max_m, max_l, sym=True, NFP=2
    )

    B_dom = B.compute_magnetic_field(coords)
    np.testing.assert_allclose(B_dom[:, 0], 0, atol=1e-13)
    np.testing.assert_allclose(B_dom[:, 1], B0 / R.flatten(), atol=1e-13)
    np.testing.assert_allclose(B_dom[:, 2], B0_Z, atol=1e-13)

    np.testing.assert_allclose(B._params["B0"], B0)

    # only nonzero coefficient of the field should be the B0 and a_ml = a_01
    np.testing.assert_allclose(B._params["B0"], B0, atol=1e-13)
    for coef, m, l in zip(B._params["a_arr"], B._params["ms"], B._params["ls"]):
        if m == 0 and l == 1:
            np.testing.assert_allclose(coef, B0_Z)
        else:
            np.testing.assert_allclose(coef, 0, atol=1e-13)
    for name in ["b_arr", "c_arr", "d_arr"]:
        np.testing.assert_allclose(B._params[name], 0, atol=1e-13)


@pytest.mark.unit
def test_domm_field_is_nonzero_and_continuous_across_Z_0():
    """Test that at Z=0 the Bphi of domm field is not discontinuous."""
    # following field should, at constant R and Z, have Bphi
    # be nonzero everywhere.
    # related to issue noted in PRs #966 and #961, where
    # before the fix in #966 was implemented the field Bphi
    # would drop to 0 discontinuously at Z=0.
    new_field = DommaschkPotentialField(1, 1, 10, 0, 0, 9, B0=0)
    Zs = np.linspace(-0.05, 0.05, 101)
    Rs = 1.1 * np.ones_like(Zs)
    phis = 0 * np.ones_like(Zs)

    Bnew = new_field.compute_magnetic_field(np.vstack([Rs, phis, Zs]).T)

    assert not np.any(np.isclose(Bnew[:, 1], 0))
